import unittest

from symbolchain.core.CodeWordsEncoder import CodeWords, CodeWordsEncoder


class CodeWordsTest(unittest.TestCase):
    def test_equality_is_supported(self):
        # Arrange:
        code_words = CodeWords(['cat', 'bird', 'dog'])

        # Act + Assert:
        self.assertEqual(code_words, CodeWords(['cat', 'bird', 'dog']))
        self.assertNotEqual(code_words, CodeWords(['cat', 'bird', 'DOG']))
        self.assertNotEqual(code_words, CodeWords(['cat', 'bird']))
        self.assertNotEqual(code_words, None)

    def test_string_is_supported(self):
        self.assertEqual('cat bird dog', str(CodeWords(['cat', 'bird', 'dog'])))

    def test_string_is_supported_custom_separator(self):
        # Arrange:
        code_words = CodeWords(['cat', 'bird', 'dog'])
        code_words.separator = '+'

        # Act + Assert:
        self.assertEqual('cat+bird+dog', str(code_words))


WORD_LIST = ['alpha', 'beta', 'gamma', 'delta', 'epsilon', 'eta', 'theta', 'iota', 'kappa', 'lambda']


class CodeWordsEncoderTest(unittest.TestCase):
    # region wordlists

    def test_cannot_create_encoder_with_less_than_two_words(self):
        # Act + Assert:
        for wordlist in [[], ['O']]:
            with self.assertRaises(ValueError):
                CodeWordsEncoder(wordlist)

    def test_can_create_encoder_with_exactly_two_words(self):
        # Arrange:
        encoder = CodeWordsEncoder(['O', 'X'])

        # Act:
        code_words = encoder.encode_int(10)
        value = encoder.decode_int(code_words)

        # Assert:
        self.assertEqual(CodeWords(['O', 'X', 'O', 'X']), code_words)
        self.assertEqual(10, value)

    def _assert_can_create_encoder_with_bip_wordlist(self, encoder, expected_words):
        # Arrange:
        value_to_encode = 3 + 1712 * 2048 + 350 * 2048 * 2048

        # Act:
        code_words = encoder.encode_int(value_to_encode)
        value = encoder.decode_int(code_words)

        # Assert:
        self.assertEqual(CodeWords(expected_words), code_words)
        self.assertEqual(value_to_encode, value)

    def test_can_create_encoder_with_default_bip_wordlist(self):
        self._assert_can_create_encoder_with_bip_wordlist(CodeWordsEncoder(), ['about', 'stock', 'cloth'])

    def test_can_create_encoder_with_custom_bip_wordlist(self):
        self._assert_can_create_encoder_with_bip_wordlist(CodeWordsEncoder('spanish'), ['abierto', 'ser', 'carro'])

    # endregion

    # region encode_int

    def _assert_can_encode_int(self, value, expected_words):
        # Arrange:
        encoder = CodeWordsEncoder(WORD_LIST)

        # Act:
        code_words = encoder.encode_int(value)

        # Assert:
        self.assertEqual(CodeWords(expected_words), code_words)

    def test_can_encode_int_zero(self):
        self._assert_can_encode_int(0, ['alpha'])

    def test_can_encode_int_to_single_word(self):
        self._assert_can_encode_int(7, ['iota'])

    def test_can_encode_int_to_multiple_words(self):
        self._assert_can_encode_int(625, ['eta', 'gamma', 'theta'])

    # endregion

    # region decode_int

    def _assert_can_decode_int(self, words, expected_value):
        # Arrange:
        encoder = CodeWordsEncoder(WORD_LIST)

        # Act:
        value = encoder.decode_int(CodeWords(words))

        # Assert:
        self.assertEqual(expected_value, value)

    def test_can_decode_int_zero(self):
        self._assert_can_decode_int(['alpha'], 0)

    def test_can_decode_int_to_single_word(self):
        self._assert_can_decode_int(['iota'], 7)

    def test_can_decode_int_to_multiple_words(self):
        self._assert_can_decode_int(['eta', 'gamma', 'theta'], 625)

    # endregion



import hashlib
from abc import abstractmethod


class Network:
    """Represents a network."""

    def __init__(self, name, identifier):
        """Creates a new network with the specified name and identifier byte."""
        self.name = name
        self.identifier = identifier

    def public_key_to_address(self, public_key):
        """Converts a public key to an address."""
        part_one_hash_builder = self.address_hasher()
        part_one_hash_builder.update(public_key.bytes)
        part_one_hash = part_one_hash_builder.digest()

        part_two_hash_builder = hashlib.new('ripemd160')
        part_two_hash_builder.update(part_one_hash)
        part_two_hash = part_two_hash_builder.digest()

        version = bytes([self.identifier]) + part_two_hash

        part_three_hash_builder = self.address_hasher()
        part_three_hash_builder.update(version)
        checksum = part_three_hash_builder.digest()[0:4]

        return self.create_address(version, checksum)

    @abstractmethod
    def address_hasher(self):
        """Gets the primary hasher to use in the public key to address conversion."""

    @abstractmethod
    def create_address(self, address_without_checksum, checksum):
        """Creates an encoded address from an address without checksum and checksum bytes."""

    def __eq__(self, other):
        return isinstance(other, Network) and self.name == other.name and self.identifier == other.identifier

    def __str__(self):
        return self.name


class NetworkLocator:
    """Provides utility functions for finding a network."""

    @staticmethod
    def find_by_name(networks, names):
        """Finds a network with a specified name within a list of networks."""
        if isinstance(names, str):
            names = [names]

        return next(network for network in networks if network.name in names)

    @staticmethod
    def find_by_identifier(networks, identifiers):
        """Finds a network with a specified identifier within a list of networks."""
        if isinstance(identifiers, int):
            identifiers = [identifiers]

        return next(network for network in networks if network.identifier in identifiers)



import base64

import sha3

from ..ByteArray import ByteArray
from ..Network import Network as BasicNetwork


class Address(ByteArray):
    """Represents a nis address."""

    SIZE = 25

    def __init__(self, address):
        """Creates an address from a decoded or encoded address."""
        raw_bytes = address
        if isinstance(address, str):
            raw_bytes = base64.b32decode(address)
        elif isinstance(address, Address):
            raw_bytes = address.bytes

        super().__init__(self.SIZE, raw_bytes, Address)

    def __str__(self):
        return base64.b32encode(self.bytes).decode('utf8')


class Network(BasicNetwork):
    """Represents a nis network."""

    def address_hasher(self):
        return sha3.keccak_256()

    def create_address(self, address_without_checksum, checksum):
        return Address(address_without_checksum + checksum)


Network.MAINNET = Network('mainnet', 0x68)
Network.TESTNET = Network('testnet', 0x98)
Network.NETWORKS = [Network.MAINNET, Network.TESTNET]



from ..BufferWriter import BufferWriter
from ..TransactionDescriptorProcessor import TransactionDescriptorProcessor
from .ImportanceTransferTransaction import ImportanceTransferTransaction
from .TransferTransaction import TransferTransaction


class TransactionFactory:
    """Factory for creating transactions."""

    def __init__(self, network, type_parsing_rules=None):
        """Creates a factory for the specified network."""
        self.network = network
        self.type_parsing_rules = type_parsing_rules

    def _create(self, name):
        transaction_classes = [ImportanceTransferTransaction, TransferTransaction]
        if name not in [transaction_class.NAME for transaction_class in transaction_classes]:
            raise ValueError('transaction named {} is not supported'.format(name))

        return next(transaction_class for transaction_class in transaction_classes if name == transaction_class.NAME)(self.network)

    def create(self, transaction_descriptor):
        """Creates a transaction from a transaction descriptor."""
        processor = TransactionDescriptorProcessor(transaction_descriptor, self.type_parsing_rules)
        transaction = self._create(processor.lookup_value('type'))

        processor.set_type_hints(type(transaction).TYPE_HINTS)
        processor.copy_to(transaction, ['type'])
        return transaction

    @staticmethod
    def attach_signature(transaction, signature):
        """Attaches a signature to a transaction."""
        writer = BufferWriter()

        transaction_buffer = transaction.serialize()
        writer.write_int(len(transaction_buffer), 4)
        writer.write_bytes(transaction_buffer)

        writer.write_int(len(signature.bytes), 4)
        writer.write_bytes(signature.bytes)
        return writer.buffer



import hashlib
import math

from mnemonic import Mnemonic


class DiceMnemonicGenerator:
    """Generates BIP39 mnemonics from dice rolls."""

    def __init__(self, num_die_sides=6, mnemonic_language='english'):
        """Creates a generator."""
        self.num_die_sides = num_die_sides
        self.mnemonic_language = mnemonic_language
        self.rolls = []

    def add_roll(self, value):
        """Adds a roll."""
        if 1 <= value <= self.num_die_sides:
            self.rolls.append(value)
            return

        raise ValueError('roll value must be between 1 and {}, inclusive'.format(self.num_die_sides))

    def frequencies(self):
        """Gets the number of times each number was rolled."""
        counts = [0] * self.num_die_sides
        for roll in self.rolls:
            counts[roll - 1] += 1

        return counts

    def to_mnemonic(self, shrink_wrap=False):
        """Generates a mnemonic by hashing the roll input. Disabling `shrink_wrap` is the same method used by coldcard wallet."""
        seed = hashlib.sha256(''.join(map(str, self.rolls)).encode('utf8')).digest()
        if not shrink_wrap:
            return (self._seed_to_mnemonic(seed), min(256, len(self.rolls) * -math.log2(1/self.num_die_sides)))

        seed_length = self._calculate_shrink_wrap_seed_length()
        return (self._seed_to_mnemonic(seed[0:seed_length]), 8 * seed_length)

    def _calculate_shrink_wrap_seed_length(self):
        min_seed_length = 16
        max_seed_length = 32

        total_roll_possibilities = self.num_die_sides ** len(self.rolls)
        if total_roll_possibilities < 2 ** (8 * min_seed_length):
            raise ValueError('shrink_wrap requires at least 128 bits of entropy but only {} bits provided'.format(total_roll_possibilities))

        seed_length = min_seed_length
        while 2 ** (8 * (seed_length + 4)) < total_roll_possibilities:
            seed_length += 4

        return min(max_seed_length, seed_length)

    def _seed_to_mnemonic(self, seed):
        return Mnemonic(self.mnemonic_language).to_mnemonic(seed)



import unittest
from binascii import hexlify

from symbol_catbuffer.AccountRestrictionFlagsDto import AccountRestrictionFlagsDto
from symbol_catbuffer.EntityTypeDto import EntityTypeDto
from symbol_catbuffer.MosaicRestrictionTypeDto import MosaicRestrictionTypeDto
from symbol_catbuffer.NetworkTypeDto import NetworkTypeDto

from symbolchain.core.CryptoTypes import PublicKey, Signature
from symbolchain.core.sym.IdGenerator import generate_mosaic_id, generate_namespace_id
from symbolchain.core.sym.Network import Address, Network
from symbolchain.core.sym.TransactionFactory import TransactionFactory
from symbolchain.tests.test.NemTestUtils import NemTestUtils

FOO_NETWORK = Network('foo', 0x54)
TEST_SIGNER_PUBLIC_KEY = NemTestUtils.randbytes(PublicKey.SIZE)


class TransactionFactoryTest(unittest.TestCase):
    # pylint: disable=too-many-public-methods

    # region create

    def _assert_transfer(self, transaction):
        self.assertEqual(0x4154, transaction.type)
        self.assertEqual(1, transaction.version)
        self.assertEqual(NetworkTypeDto.PUBLIC_TEST, transaction.network)

    def _assert_can_create_known_transaction_from_descriptor(self, create_function_accessor):
        # Arrange:
        factory = TransactionFactory(Network.PUBLIC_TEST)

        # Act:
        transaction = create_function_accessor(factory)({
            'type': 'transfer',
            'signer_public_key': TEST_SIGNER_PUBLIC_KEY
        })

        # Assert:
        self._assert_transfer(transaction)
        self.assertEqual(TEST_SIGNER_PUBLIC_KEY, transaction.signer_public_key)

    def test_can_create_known_transaction_from_descriptor(self):
        self._assert_can_create_known_transaction_from_descriptor(lambda factory: factory.create)

    def test_can_create_known_transaction_from_descriptor_embedded(self):
        self._assert_can_create_known_transaction_from_descriptor(lambda factory: factory.create_embedded)

    def _assert_cannot_create_unknown_transaction_from_descriptor(self, create_function_accessor):
        # Arrange:
        factory = TransactionFactory(Network.PUBLIC_TEST)

        # Act + Assert:
        with self.assertRaises(ValueError):
            create_function_accessor(factory)({
                'type': 'xtransfer',
                'signer_public_key': TEST_SIGNER_PUBLIC_KEY
            })

    def test_cannot_create_unknown_transaction_from_descriptor(self):
        self._assert_cannot_create_unknown_transaction_from_descriptor(lambda factory: factory.create)

    def test_cannot_create_unknown_transaction_from_descriptor_embedded(self):
        self._assert_cannot_create_unknown_transaction_from_descriptor(lambda factory: factory.create_embedded)

    def _assert_can_create_known_transaction_with_multiple_overrides(self, create_function_accessor):
        # Arrange:
        factory = TransactionFactory(Network.PUBLIC_TEST, {
            Address: lambda address: address + ' ADDRESS',
            PublicKey: lambda address: address + ' PUBLICKEY'
        })

        # Act:
        transaction = create_function_accessor(factory)({
            'type': 'transfer',
            'signer_public_key': 'signer_name',
            'recipient_address': 'recipient_name',
            'message': 'hello world',
            'mosaics': [(0x12345678ABCDEF, 12345)]
        })

        # Assert:
        self._assert_transfer(transaction)
        self.assertEqual('signer_name PUBLICKEY', transaction.signer_public_key)

        self.assertEqual('recipient_name ADDRESS', transaction.recipient_address)
        self.assertEqual(b'hello world', transaction.message)
        self.assertEqual([(0x12345678ABCDEF, 12345)], transaction.mosaics)

    def test_can_create_known_transaction_with_multiple_overrides(self):
        self._assert_can_create_known_transaction_with_multiple_overrides(lambda factory: factory.create)

    def test_can_create_known_transaction_with_multiple_overrides_embedded(self):
        self._assert_can_create_known_transaction_with_multiple_overrides(lambda factory: factory.create_embedded)

    # endregion

    # region flags and enums handling

    def _assert_can_create_transaction_with_default_flags_handling(self, create_function_accessor):
        # Arrange:
        factory = TransactionFactory(Network.PUBLIC_TEST)

        # Act:
        transaction = create_function_accessor(factory)({
            'type': 'accountAddressRestriction',
            'signer_public_key': 'signer_name',
            'restriction_flags': 'address mosaic_id transaction_type outgoing block'
        })

        # Assert:
        Flags = AccountRestrictionFlagsDto
        self.assertEqual(
            [Flags.ADDRESS, Flags.MOSAIC_ID, Flags.TRANSACTION_TYPE, Flags.OUTGOING, Flags.BLOCK],
            transaction.restriction_flags)

    def test_can_create_transaction_with_default_flags_handling(self):
        self._assert_can_create_transaction_with_default_flags_handling(lambda factory: factory.create)

    def test_can_create_transaction_with_default_flags_handling_embedded(self):
        self._assert_can_create_transaction_with_default_flags_handling(lambda factory: factory.create_embedded)

    def _assert_can_create_transaction_with_default_enum_handling(self, create_function_accessor):
        # Arrange:
        factory = TransactionFactory(Network.PUBLIC_TEST)

        # Act:
        transaction = create_function_accessor(factory)({
            'type': 'mosaicGlobalRestriction',
            'signer_public_key': 'signer_name',
            'mosaic_id': 0x12345,
            'previous_restriction_type': 'le',
            'new_restriction_type': 'gt'
        })

        # Assert:
        self.assertEqual(MosaicRestrictionTypeDto.LE, transaction.previous_restriction_type)
        self.assertEqual(MosaicRestrictionTypeDto.GT, transaction.new_restriction_type)

    def test_can_create_transaction_with_default_enum_handling(self):
        self._assert_can_create_transaction_with_default_enum_handling(lambda factory: factory.create)

    def test_can_create_transaction_with_default_enum_handling_embedded(self):
        self._assert_can_create_transaction_with_default_enum_handling(lambda factory: factory.create_embedded)

    def _assert_can_create_transaction_with_default_enum_array_handling(self, create_function_accessor):
        # Arrange:
        factory = TransactionFactory(Network.PUBLIC_TEST)

        # Act:
        transaction = create_function_accessor(factory)({
            'type': 'accountOperationRestriction',
            'signer_public_key': 'signer_name',
            'restriction_additions': [
                'transfer_transaction',
                'account_key_link_transaction'
            ],
            'restriction_deletions': [
                'vrf_key_link_transaction',
                'voting_key_link_transaction'
            ]
        })

        # Assert:
        expected_additions = [EntityTypeDto.TRANSFER_TRANSACTION, EntityTypeDto.ACCOUNT_KEY_LINK_TRANSACTION]
        self.assertEqual(expected_additions, transaction.restriction_additions)
        expected_deletions = [EntityTypeDto.VRF_KEY_LINK_TRANSACTION, EntityTypeDto.VOTING_KEY_LINK_TRANSACTION]
        self.assertEqual(expected_deletions, transaction.restriction_deletions)

    def test_can_create_transaction_with_default_enum_array_handling(self):
        self._assert_can_create_transaction_with_default_enum_array_handling(lambda factory: factory.create)

    def test_can_create_transaction_with_default_enum_array_handling_embedded(self):
        self._assert_can_create_transaction_with_default_enum_array_handling(lambda factory: factory.create_embedded)

    # endregion

    # region id autogeneration

    def _assert_can_autogenerate_namespace_registration_root_id(self, create_function_accessor):
        # Arrange:
        factory = TransactionFactory(Network.PUBLIC_TEST)

        # Act:
        transaction = create_function_accessor(factory)({
            'type': 'namespaceRegistration',
            'signer_public_key': TEST_SIGNER_PUBLIC_KEY,
            'registration_type': 'root',
            'duration': 123,
            'name': 'roger'
        })

        # Assert:
        expected_id = generate_namespace_id('roger')
        self.assertEqual(expected_id, transaction.id)

    def test_can_autogenerate_namespace_registration_root_id(self):
        self._assert_can_autogenerate_namespace_registration_root_id(lambda factory: factory.create)

    def test_can_autogenerate_namespace_registration_root_id_embedded(self):
        self._assert_can_autogenerate_namespace_registration_root_id(lambda factory: factory.create_embedded)

    def _assert_can_autogenerate_namespace_registration_child_id(self, create_function_accessor):
        # Arrange:
        factory = TransactionFactory(Network.PUBLIC_TEST)

        # Act:
        transaction = create_function_accessor(factory)({
            'type': 'namespaceRegistration',
            'signer_public_key': TEST_SIGNER_PUBLIC_KEY,
            'registration_type': 'child',
            'parent_id': generate_namespace_id('roger'),
            'name': 'charlie'
        })

        # Assert:
        expected_id = generate_namespace_id('charlie', generate_namespace_id('roger'))
        self.assertEqual(expected_id, transaction.id)

    def test_can_autogenerate_namespace_registration_child_id(self):
        self._assert_can_autogenerate_namespace_registration_child_id(lambda factory: factory.create)

    def test_can_autogenerate_namespace_registration_child_id_embedded(self):
        self._assert_can_autogenerate_namespace_registration_child_id(lambda factory: factory.create_embedded)

    def _assert_can_autogenerate_mosaic_definition_id(self, create_function_accessor):
        # Arrange:
        factory = TransactionFactory(Network.PUBLIC_TEST)

        # Act:
        transaction = create_function_accessor(factory)({
            'type': 'mosaicDefinition',
            'signer_public_key': TEST_SIGNER_PUBLIC_KEY,
            'duration': 1,
            'nonce': 123,
            'flags': 'transferable restrictable',
            'divisibility': 2
        })

        # Assert:
        expected_id = generate_mosaic_id(factory.network.public_key_to_address(PublicKey(TEST_SIGNER_PUBLIC_KEY)), 123)
        self.assertEqual(expected_id, transaction.id)

    def test_can_autogenerate_mosaic_definition_id(self):
        self._assert_can_autogenerate_mosaic_definition_id(lambda factory: factory.create)

    def test_can_autogenerate_mosaic_definition_id_embedded(self):
        self._assert_can_autogenerate_mosaic_definition_id(lambda factory: factory.create_embedded)

    # endregion

    # region attach_signature

    def test_can_attach_signature_to_transaction(self):
        # Arrange:
        factory = TransactionFactory(Network.PUBLIC_TEST)
        transaction = factory.create({
            'type': 'transfer',
            'signer_public_key': TEST_SIGNER_PUBLIC_KEY
        })
        signature = NemTestUtils.randcryptotype(Signature)

        # Sanity:
        self.assertEqual(Signature.zero().bytes, transaction.signature)

        # Act:
        signed_transaction_buffer = factory.attach_signature(transaction, signature)

        # Assert:
        self._assert_transfer(transaction)
        self.assertEqual(TEST_SIGNER_PUBLIC_KEY, transaction.signer_public_key)

        self.assertEqual(signature.bytes, transaction.signature)

        serialized_transaction_hex = hexlify(transaction.serialize()).decode('utf8').upper()
        expected_buffer = '{{"payload": "{}"}}'.format(serialized_transaction_hex).encode('utf8')
        self.assertEqual(expected_buffer, signed_transaction_buffer)

    # endregion



import unittest
from functools import reduce

from symbolchain.core.CryptoTypes import PublicKey, Signature
from symbolchain.core.nis1.Network import Address, Network
from symbolchain.core.nis1.TransactionFactory import TransactionFactory
from symbolchain.tests.test.NemTestUtils import NemTestUtils

FOO_NETWORK = Network('foo', 0x54)


class MockTransaction:
    def __init__(self, buffer):
        self.buffer = buffer

    def serialize(self):
        return self.buffer


class TransactionFactoryTest(unittest.TestCase):
    # region create

    def test_can_create_known_transaction_from_descriptor(self):
        # Arrange:
        for transaction_name, transaction_type in [('transfer', 0x0101), ('importance-transfer', 0x0801)]:
            factory = TransactionFactory(FOO_NETWORK)

            # Act:
            transaction = factory.create({'type': transaction_name})

            # Assert:
            self.assertEqual(transaction_type, transaction.type)
            self.assertEqual(0x54000001, transaction.version)

    def test_cannot_create_unknown_transaction_from_descriptor(self):
        # Arrange:
        factory = TransactionFactory(FOO_NETWORK)

        # Act + Assert:
        with self.assertRaises(ValueError):
            factory.create({'type': 'multisig'})

    def test_can_create_known_transaction_with_multiple_overrides(self):
        # Arrange:
        factory = TransactionFactory(FOO_NETWORK, {
            Address: lambda address: address + ' ADDRESS',
            PublicKey: lambda address: address + ' PUBLICKEY'
        })

        # Act:
        transaction = factory.create({
            'type': 'transfer',
            'timestamp': 98765,
            'signer_public_key': 'signer_name',
            'recipient_address': 'recipient_name',
            'message': 'hello world',
        })

        # Assert:
        self.assertEqual(0x0101, transaction.type)
        self.assertEqual(0x54000001, transaction.version)
        self.assertEqual(98765, transaction.timestamp)
        self.assertEqual('signer_name PUBLICKEY', transaction.signer_public_key)

        self.assertEqual('recipient_name ADDRESS', transaction.recipient_address)
        self.assertEqual(b'hello world', transaction.message)

    # endregion

    # region attach_signature

    def test_can_attach_signature_to_transaction(self):
        # Arrange:
        transaction = MockTransaction(bytes([0x44, 0x55, 0x98, 0x12, 0x71, 0xAB, 0x72]))
        signature = NemTestUtils.randcryptotype(Signature)

        # Act:
        signed_transaction_buffer = TransactionFactory.attach_signature(transaction, signature)

        # Assert:
        expected_buffers = [
            [0x07, 0x00, 0x00, 0x00],  # transaction length
            [0x44, 0x55, 0x98, 0x12, 0x71, 0xAB, 0x72],  # transaction
            [0x40, 0x00, 0x00, 0x00],  # signature length
            signature.bytes  # signature
        ]
        expected_buffer = reduce(lambda x, y: bytes(x) + bytes(y), expected_buffers)
        self.assertEqual(expected_buffer, signed_transaction_buffer)

    # endregion



import random


class NemTestUtils:
    """NEM common test utilities."""

    @staticmethod
    def randbytes(count):
        """Generates random bytes. Polyfill for random.randbytes in python 3.9."""
        if hasattr(random, 'randbytes'):
            # pylint: disable=no-member
            return random.randbytes(count)

        return bytes([random.randint(0x00, 0xFF) for _ in range(0, count)])

    @staticmethod
    def randcryptotype(crypto_type_class):
        """Generates a random crypto type."""
        return crypto_type_class(NemTestUtils.randbytes(crypto_type_class.SIZE))



import os
import tempfile
import unittest

import qrcode

from symbolchain.core.QrStorage import QrStorage
from symbolchain.tests.test.NemTestUtils import NemTestUtils

MAX_DATA_BYTES = 1155


def write_random_qrcode(directory, name, data_size):
    QrStorage(directory).save(name, NemTestUtils.randbytes(data_size))


class QrStorageTest(unittest.TestCase):
    # region save

    def _assert_can_save_buffer(self, data_size):
        # Arrange:
        with tempfile.TemporaryDirectory() as temp_directory:
            storage = QrStorage(temp_directory)

            buffer = NemTestUtils.randbytes(data_size)

            # Act:
            storage.save('foo', buffer)

            # Assert:
            self.assertEqual(1, len(os.listdir(temp_directory)))
            self.assertTrue(os.path.exists(os.path.join(temp_directory, 'foo.png')))

    def test_can_save_empty_buffer(self):
        self._assert_can_save_buffer(0)

    def test_can_save_nonempty_buffer(self):
        self._assert_can_save_buffer(100)

    def test_can_save_max_buffer(self):
        self._assert_can_save_buffer(MAX_DATA_BYTES)

    def test_cannot_save_more_than_max_buffer(self):
        # Arrange:
        with tempfile.TemporaryDirectory() as temp_directory:
            storage = QrStorage(temp_directory)

            buffer = NemTestUtils.randbytes(MAX_DATA_BYTES + 1)

            # Act + Assert:
            with self.assertRaises(qrcode.exceptions.DataOverflowError):
                storage.save('foo', buffer)

    # endregion

    # region load

    def _assert_can_roundtrip_buffer(self, data_size):
        # Arrange:
        with tempfile.TemporaryDirectory() as temp_directory:
            storage = QrStorage(temp_directory)

            buffer = NemTestUtils.randbytes(data_size)
            storage.save('foo', buffer)

            # Act:
            loaded_buffer = storage.load('foo')

            # Assert:
            self.assertEqual(buffer, loaded_buffer)

    def test_can_roundtrip_empty_buffer(self):
        self._assert_can_roundtrip_buffer(0)

    def test_can_roundtrip_nonempty_buffer(self):
        self._assert_can_roundtrip_buffer(100)

    def test_cannot_load_qrcode_that_does_not_exist(self):
        # Arrange:
        with tempfile.TemporaryDirectory() as temp_directory:
            storage = QrStorage(temp_directory)

            buffer = NemTestUtils.randbytes(100)
            storage.save('foo', buffer)

            # Sanity:
            self.assertEqual(1, len(os.listdir(temp_directory)))

            # Act + Assert:
            with self.assertRaises(FileNotFoundError):
                storage.load('bar')

    # endregion



import unittest
from functools import reduce

from symbolchain.core.CryptoTypes import PublicKey
from symbolchain.core.nis1.ImportanceTransferTransaction import ImportanceTransferTransaction
from symbolchain.core.nis1.Network import Network
from symbolchain.tests.test.BasicNisTransactionTest import BasicNisTransactionTest, NisTransactionTestDescriptor

FOO_NETWORK = Network('foo', 0x54)


class ImportanceTransferTransactionTest(BasicNisTransactionTest, unittest.TestCase):
    # region basic

    @staticmethod
    def get_test_descriptor():
        return NisTransactionTestDescriptor(ImportanceTransferTransaction, 'importance-transfer', 0x0801, FOO_NETWORK)

    def test_can_create(self):
        # Act:
        transaction = ImportanceTransferTransaction(FOO_NETWORK)

        # Assert:
        self.assertEqual(0x0801, transaction.type)
        self.assertEqual(0x54000001, transaction.version)
        self.assertEqual(0, transaction.timestamp)
        self.assertEqual(None, transaction.signer_public_key)

        self.assertEqual(0, transaction.mode)
        self.assertEqual(None, transaction.remote_account_public_key)

        # - properties
        self.assertEqual(150000, transaction.fee)
        self.assertEqual(60 * 60, transaction.deadline)

    # endregion

    # region serialize

    @staticmethod
    def _create_transfer_for_serialization_tests():
        transaction = ImportanceTransferTransaction(FOO_NETWORK)
        transaction.timestamp = 12345
        transaction.signer_public_key = PublicKey('D6C3845431236C5A5A907A9E45BD60DA0E12EFD350B970E7F58E3499E2E7A2F0')

        transaction.mode = 2
        transaction.remote_account_public_key = PublicKey('9764026AA71A3CD0189990D1B7B8275B8D80863CF271235DFC745F30651E93AA')
        return transaction

    def test_can_serialize(self):
        # Arrange:
        transaction = self._create_transfer_for_serialization_tests()

        # Act:
        buffer = transaction.serialize()

        # Assert:
        expected_buffers = [
            [0x01, 0x08, 0x00, 0x00],  # type
            [0x01, 0x00, 0x00, 0x54],  # version
            [0x39, 0x30, 0x00, 0x00],  # timestamp
            [0x20, 0x00, 0x00, 0x00],  # public key length
            transaction.signer_public_key.bytes,  # signer public key
            [0xF0, 0x49, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00],  # fee
            [0x49, 0x3E, 0x00, 0x00],  # deadline
            [0x02, 0x00, 0x00, 0x00],  # mode
            [0x20, 0x00, 0x00, 0x00],  # public key length
            transaction.remote_account_public_key.bytes,  # remote account public key
        ]
        expected_buffer = reduce(lambda x, y: bytes(x) + bytes(y), expected_buffers)
        self.assertEqual(expected_buffer, buffer)

    # endregion

    # region string

    def test_can_create_string_representation(self):
        # Arrange:
        transaction = self._create_transfer_for_serialization_tests()

        # Act:
        transaction_str = str(transaction)

        # Assert:
        expected_transaction_str = '\n'.join([
            '                     type = 2049 [0x801]',
            '                  version = 1409286145 [0x54000001]',
            '                timestamp = 12345 [0x3039]',
            '        signer_public_key = D6C3845431236C5A5A907A9E45BD60DA0E12EFD350B970E7F58E3499E2E7A2F0',
            '                      fee = 150000 [0x249F0]',
            '                 deadline = 15945 [0x3E49]',
            '                     mode = 2 [0x2]',
            'remote_account_public_key = 9764026AA71A3CD0189990D1B7B8275B8D80863CF271235DFC745F30651E93AA'
        ])
        self.assertEqual(expected_transaction_str, transaction_str)

    # endregion



import sha3
from symbol_catbuffer.EntityTypeDto import EntityTypeDto

from ..CryptoTypes import Hash256, PublicKey, Signature
from ..Network import NetworkLocator
from ..sym.KeyPair import KeyPair, Verifier
from ..sym.Network import Address, Network
from ..sym.TransactionFactory import TransactionFactory

TRANSACTION_HEADER_SIZE = sum(field[1] for field in [
    ('size', 4),
    ('reserved1', 4),
    ('signature', Signature.SIZE),
    ('signer', PublicKey.SIZE),
    ('reserved2', 4)
])

AGGREGATE_HASHED_SIZE = sum(field[1] for field in [
    ('version_network_type', 4),
    ('max_fee', 8),
    ('deadline', 8),
    ('transactions_hash', Hash256.SIZE)
])


class SymFacade:
    """Facade used to interact with symbol blockchain."""

    BIP32_COIN_ID = 4343
    BIP32_CURVE_NAME = 'ed25519'

    KeyPair = KeyPair
    Verifier = Verifier

    def __init__(self, sym_network_name, account_descriptor_repository=None):
        """Creates a symbol facade."""
        self.network = NetworkLocator.find_by_name(Network.NETWORKS, sym_network_name)
        self.account_descriptor_repository = account_descriptor_repository
        self.transaction_factory = self._create_sym_transaction_factory()

    def _create_sym_transaction_factory(self):
        type_parsing_rules = None
        if self.account_descriptor_repository:
            type_parsing_rules = self.account_descriptor_repository.to_type_parsing_rules_map(self._create_sym_type_to_property_mapping())

            for key in type_parsing_rules.keys():
                type_parsing_rules[key] = self._unwrap_mapped_type(type_parsing_rules[key])

        return TransactionFactory(self.network, type_parsing_rules)

    @staticmethod
    def _create_sym_type_to_property_mapping():
        return {
            Address: 'address',
            PublicKey: 'public_key'
        }

    @staticmethod
    def _unwrap_mapped_type(transform):
        return lambda value: transform(value).bytes

    def hash_transaction(self, transaction):
        """Hashes a symbol transaction."""
        hasher = sha3.sha3_256()
        hasher.update(transaction.signature)
        hasher.update(transaction.signer_public_key)
        hasher.update(self.network.generation_hash_seed.bytes)
        hasher.update(self._transaction_data_buffer(transaction.serialize()))
        return Hash256(hasher.digest())

    def sign_transaction(self, key_pair, transaction):
        """Signs a symbol transaction."""
        sign_buffer = self.network.generation_hash_seed.bytes
        sign_buffer += self._transaction_data_buffer(transaction.serialize())
        return key_pair.sign(sign_buffer)

    def verify_transaction(self, transaction, signature):
        """Verifies a symbol transaction."""
        verify_buffer = self.network.generation_hash_seed.bytes
        verify_buffer += self._transaction_data_buffer(transaction.serialize())
        return Verifier(PublicKey(transaction.signer_public_key)).verify(verify_buffer, signature)

    @staticmethod
    def bip32_node_to_key_pair(bip32_node):
        """Derives a symbol KeyPair from a BIP32 node."""
        return KeyPair(bip32_node.private_key)

    @staticmethod
    def _transaction_data_buffer(transaction_buffer):
        data_buffer_start = TRANSACTION_HEADER_SIZE
        data_buffer_end = len(transaction_buffer)
        if SymFacade._is_aggregate_transaction(transaction_buffer):
            data_buffer_end = TRANSACTION_HEADER_SIZE + AGGREGATE_HASHED_SIZE

        return transaction_buffer[data_buffer_start:data_buffer_end]

    @staticmethod
    def _is_aggregate_transaction(transaction_buffer):
        transaction_type_offset = TRANSACTION_HEADER_SIZE + 2  # skip version and network byte
        transaction_type = (transaction_buffer[transaction_type_offset + 1] << 8) + transaction_buffer[transaction_type_offset]
        aggregate_types = [EntityTypeDto.AGGREGATE_BONDED_TRANSACTION.value, EntityTypeDto.AGGREGATE_COMPLETE_TRANSACTION.value]
        return transaction_type in aggregate_types



import os
import tempfile
import unittest

from symbolchain.core.AccountDescriptorRepository import AccountDescriptorRepository
from symbolchain.core.CryptoTypes import Hash256, PrivateKey, PublicKey, Signature
from symbolchain.core.facade.BatchOperations import BatchOperations
from symbolchain.core.facade.NisFacade import NisFacade
from symbolchain.core.nis1.Network import Address
from symbolchain.core.PrivateKeyStorage import PrivateKeyStorage
from symbolchain.core.QrSignatureStorage import QrSignatureStorage
from symbolchain.tests.test.NemTestUtils import NemTestUtils

# NisFacade is used in tests, so constants are using nis formats

TEST_PUBLIC_KEY = PublicKey('0F74A2F537CD9C986DF018994DDE75BDEEE05E35EB9FE27ADF506CA8475064F7')
TEST_PRIVATE_KEY = PrivateKey('D9639DC6F49DAD02A42FD8C217F1B1B4F8CE31CCD770388B645E639C72FF24FA')
ALICE_ADDRESS = Address('TALIC33PNVKIMNXVOCOQGWLZK52K4XALZBNE2ISF')
BOB_ADDRESS = Address('TBZ6JK5YOCU6UPSSZ5D3G27UHAPHTY5HDQCDS5YA')

ACCOUNTS_YAML_INPUT = '''
- public_key: {test_public_key}
  name: TEST

- address: {alice_address}
  name: ALICE
'''.format(test_public_key=TEST_PUBLIC_KEY, alice_address=ALICE_ADDRESS)

TRANSACTIONS_YAML_INPUT = '''
- type: transfer
  signer_public_key: TEST
  recipient_address: ALICE
  amount: 3000000
  message: Hello world!

- type: transfer
  signer_public_key: TEST
  recipient_address: {bob_address}
  amount: 1000000
'''.format(bob_address=BOB_ADDRESS)


class BatchOperationsTest(unittest.TestCase):
    # region load_all

    def test_can_load_all_transactions(self):
        # Arrange:
        operations = self._create_operations()

        # Act:
        transactions = operations.load_all(TRANSACTIONS_YAML_INPUT)

        # Assert:
        self.assertEqual(2, len(transactions))

        self.assertEqual(0x0101, transactions[0].type)
        self.assertEqual(0x98000001, transactions[0].version)
        self.assertEqual(TEST_PUBLIC_KEY, transactions[0].signer_public_key)
        self.assertEqual(ALICE_ADDRESS, transactions[0].recipient_address)
        self.assertEqual(3000000, transactions[0].amount)
        self.assertEqual(b'Hello world!', transactions[0].message)

        self.assertEqual(0x0101, transactions[1].type)
        self.assertEqual(0x98000001, transactions[1].version)
        self.assertEqual(TEST_PUBLIC_KEY, transactions[1].signer_public_key)
        self.assertEqual(BOB_ADDRESS, transactions[1].recipient_address)
        self.assertEqual(1000000, transactions[1].amount)
        self.assertEqual(None, transactions[1].message)

    # endregion

    # region sign_all

    def test_cannot_sign_all_transactions_when_private_keys_are_not_available(self):
        # Arrange:
        operations = self._create_operations('test')
        transactions = operations.load_all(TRANSACTIONS_YAML_INPUT)

        with tempfile.TemporaryDirectory() as temp_directory:
            (private_key_storage, signature_storage) = self._create_storages(temp_directory, False)

            # Act + Assert:
            with self.assertRaises(FileNotFoundError):
                operations.sign_all(transactions, private_key_storage, signature_storage)

    def test_can_sign_all_transactions_when_private_keys_are_available(self):
        # Arrange:
        operations = self._create_operations('test')
        transactions = operations.load_all(TRANSACTIONS_YAML_INPUT)

        with tempfile.TemporaryDirectory() as temp_directory:
            (private_key_storage, signature_storage) = self._create_storages(temp_directory)

            # Act:
            operations.sign_all(transactions, private_key_storage, signature_storage)

            # Assert:
            signature_filenames = os.listdir(os.path.join(temp_directory, 'signatures'))
            self.assertEqual(2, len(signature_filenames))
            for filename in ['sig_test0.png', 'sig_test1.png']:
                self.assertTrue(filename in signature_filenames)

    # endregion

    # region prepare_all

    def test_cannot_prepare_all_when_more_transactions_than_signature_groups(self):
        # Arrange:
        operations = self._create_operations('test')
        transactions = operations.load_all(TRANSACTIONS_YAML_INPUT)

        with tempfile.TemporaryDirectory() as temp_directory:
            (private_key_storage, signature_storage) = self._create_storages(temp_directory)
            operations.sign_all(transactions, private_key_storage, signature_storage)

            payload_directory = os.path.join(temp_directory, 'payloads')
            os.mkdir(payload_directory)

            # Act + Assert:
            with self.assertRaises(FileNotFoundError):
                operations.prepare_all(transactions + [transactions[0]], signature_storage, payload_directory)

            # Sanity: no payloads were created
            self.assertEqual(0, len(os.listdir(payload_directory)))

    def _assert_cannot_prepare_all_when_transaction_verification_fails(self, corrupt_hash=False, corrupt_signature=False):
        # Arrange:
        operations = self._create_operations('test')
        transactions = operations.load_all(TRANSACTIONS_YAML_INPUT)

        with tempfile.TemporaryDirectory() as temp_directory:
            (private_key_storage, signature_storage) = self._create_storages(temp_directory)
            operations.sign_all(transactions, private_key_storage, signature_storage)

            payload_directory = os.path.join(temp_directory, 'payloads')
            os.mkdir(payload_directory)

            # overrwrite second qrcode with the intention to corrupt it
            if corrupt_hash:
                transaction_hash = NemTestUtils.randcryptotype(Hash256)
            else:
                transaction_hash = operations.facade.hash_transaction(transactions[1])

            if corrupt_signature:
                signature = NemTestUtils.randcryptotype(Signature)
            else:
                signer_public_key = transactions[1].signer_public_key
                signer_account_name = operations.facade.account_descriptor_repository.find_by_public_key(signer_public_key).name
                signer_private_key = private_key_storage.load(signer_account_name)
                signature = operations.facade.KeyPair(signer_private_key).sign(transactions[1].serialize())

            signature_storage.save('sig_test1', transaction_hash, [signature])

            # Act + Assert:
            with self.assertRaises(BatchOperations.PrepareError):
                operations.prepare_all(transactions, signature_storage, payload_directory)

            # Sanity: no payloads were created
            self.assertEqual(0, len(os.listdir(payload_directory)))

    def test_cannot_prepare_all_when_transaction_hash_does_not_match(self):
        self._assert_cannot_prepare_all_when_transaction_verification_fails(corrupt_hash=True)

    def test_cannot_prepare_all_when_transaction_signature_does_not_match(self):
        self._assert_cannot_prepare_all_when_transaction_verification_fails(corrupt_signature=True)

    def test_can_prepare_all_when_all_transactions_match_signatures(self):
        # Arrange:
        operations = self._create_operations('test')
        transactions = operations.load_all(TRANSACTIONS_YAML_INPUT)

        with tempfile.TemporaryDirectory() as temp_directory:
            (private_key_storage, signature_storage) = self._create_storages(temp_directory)
            operations.sign_all(transactions, private_key_storage, signature_storage)

            payload_directory = os.path.join(temp_directory, 'payloads')
            os.mkdir(payload_directory)

            # Act:
            operations.prepare_all(transactions, signature_storage, payload_directory)

            # Assert: correct size payloads were written
            payload_filenames = os.listdir(payload_directory)
            self.assertEqual(2, len(payload_filenames))
            for i, filename in enumerate(['payload_test0.dat', 'payload_test1.dat']):
                self.assertTrue(filename in payload_filenames)

                expected_file_size = len(transactions[i].serialize()) + Signature.SIZE + 8
                actual_file_size = os.path.getsize(os.path.join(payload_directory, filename))
                self.assertEqual(expected_file_size, actual_file_size)

    # endregion

    @staticmethod
    def _create_operations(output_file_prefix=''):
        facade = NisFacade('testnet', AccountDescriptorRepository(ACCOUNTS_YAML_INPUT))
        return BatchOperations(facade, output_file_prefix)

    @staticmethod
    def _create_storages(temp_directory, save_test_private_key=True):
        private_key_directory = os.path.join(temp_directory, 'private_keys')
        os.mkdir(private_key_directory)
        private_key_storage = PrivateKeyStorage(private_key_directory, 'password')
        if save_test_private_key:
            private_key_storage.save('TEST', TEST_PRIVATE_KEY)

        signature_directory = os.path.join(temp_directory, 'signatures')
        os.mkdir(signature_directory)
        signature_storage = QrSignatureStorage(signature_directory)

        return (private_key_storage, signature_storage)



import unittest

from symbolchain.core.AccountDescriptorRepository import AccountDescriptorRepository
from symbolchain.core.Bip32 import Bip32
from symbolchain.core.CryptoTypes import Hash256, PrivateKey, PublicKey, Signature
from symbolchain.core.facade.NisFacade import NisFacade
from symbolchain.tests.test.NemTestUtils import NemTestUtils

YAML_INPUT = '''
- public_key: A59277D56E9F4FA46854F5EFAAA253B09F8AE69A473565E01FD9E6A738E4AB74
  name: TEST
'''


class NisFacadeTest(unittest.TestCase):
    # region constants

    def test_bip32_constants_are_correct(self):
        self.assertEqual(43, NisFacade.BIP32_COIN_ID)
        self.assertEqual('ed25519-keccak', NisFacade.BIP32_CURVE_NAME)

    def test_key_pair_is_correct(self):
        # Arrange:
        private_key = PrivateKey('ED4C70D78104EB11BCD73EBDC512FEBC8FBCEB36A370C957FF7E266230BB5D57')  # reversed

        # Act:
        key_pair = NisFacade.KeyPair(private_key)

        # Assert:
        self.assertEqual(PublicKey('D6C3845431236C5A5A907A9E45BD60DA0E12EFD350B970E7F58E3499E2E7A2F0'), key_pair.public_key)

    def test_can_sign_and_verify(self):
        # Arrange:
        private_key = PrivateKey.random()
        key_pair = NisFacade.KeyPair(private_key)
        message = NemTestUtils.randbytes(21)

        # Act:
        signature = key_pair.sign(message)
        is_verified = NisFacade.Verifier(key_pair.public_key).verify(message, signature)

        # Assert:
        self.assertTrue(is_verified)

    # endregion

    # region constructor

    def test_can_create_around_known_network(self):
        # Act:
        facade = NisFacade('testnet')
        transaction = facade.transaction_factory.create({'type': 'transfer'})

        # Assert:
        self.assertEqual('testnet', facade.network.name)

        self.assertEqual(0x0101, transaction.type)
        self.assertEqual(0x98000001, transaction.version)

    def test_cannot_create_around_unknown_network(self):
        # Act:
        with self.assertRaises(StopIteration):
            NisFacade('foo')

    # endregion

    # region hash_transaction / sign_transaction

    @staticmethod
    def _create_real_transfer():
        facade = NisFacade('testnet', AccountDescriptorRepository(YAML_INPUT))
        transaction = facade.transaction_factory.create({
            'type': 'transfer',
            'timestamp': 187751734,
            'signer_public_key': 'TEST',
            'recipient_address': 'TALIC33PNVKIMNXVOCOQGWLZK52K4XALZBNE2ISF',
            'amount': 1000000
        })
        return transaction

    def test_can_hash_transaction(self):
        # Arrange:
        transaction = self._create_real_transfer()

        # Act:
        hash_value = NisFacade.hash_transaction(transaction)

        # Assert:
        self.assertEqual(Hash256('520547F6B7E8D7217AB5D8DB9E572682B05C6BAA7F3F900DE1429CEC43253543'), hash_value)

    def test_can_sign_transaction(self):
        # Arrange:
        private_key = PrivateKey('EDB671EB741BD676969D8A035271D1EE5E75DF33278083D877F23615EB839FEC')
        transaction = self._create_real_transfer()

        # Act:
        signature = NisFacade.sign_transaction(NisFacade.KeyPair(private_key), transaction)

        # Assert:
        expected_signature = Signature(''.join([
            'AD1BD313FDD7F8502FAD22854569A515F87E50EAB01F1CCC60D93FBDFC9AFF2E'
            '89D7EF3C045AFED3C19827BBBF70F4DD12541189CE4F946CC0CB13B15558EC0C'
        ]))
        self.assertEqual(expected_signature, signature)

    def test_can_verify_transaction(self):
        # Arrange:
        private_key = PrivateKey('EDB671EB741BD676969D8A035271D1EE5E75DF33278083D877F23615EB839FEC')
        transaction = self._create_real_transfer()

        # Act:
        signature = NisFacade.sign_transaction(NisFacade.KeyPair(private_key), transaction)
        is_verified = NisFacade.verify_transaction(transaction, signature)

        # Assert:
        self.assertTrue(is_verified)

    # endregion

    # region bip32_node_to_key_pair

    def _assert_bip32_child_public_keys(self, passphrase, expected_child_public_keys):
        # Arrange:
        mnemonic_seed = ' '.join([
            'hamster', 'diagram', 'private', 'dutch', 'cause', 'delay', 'private', 'meat', 'slide', 'toddler', 'razor', 'book',
            'happy', 'fancy', 'gospel', 'tennis', 'maple', 'dilemma', 'loan', 'word', 'shrug', 'inflict', 'delay', 'length'
        ])

        # Act:
        root_node = Bip32(NisFacade.BIP32_CURVE_NAME).from_mnemonic(mnemonic_seed, passphrase)

        child_public_keys = []
        for i in range(0, len(expected_child_public_keys)):
            child_node = root_node.derive_path([44, NisFacade.BIP32_COIN_ID, i, 0, 0])
            child_key_pair = NisFacade.bip32_node_to_key_pair(child_node)
            child_public_keys.append(child_key_pair.public_key)

        # Assert:
        self.assertEqual(expected_child_public_keys, child_public_keys)

    def test_can_use_bip32_derivation_without_passphrase(self):
        self._assert_bip32_child_public_keys('', [
            PublicKey('6C42BFAD2199CCB5C64E59868CC7A3F2AD29BDDCEB9754157DF136535E6B5EBA'),
            PublicKey('782FF2375F75524106356092B4EE4BA098D28CF6571F1643867B9A11AEF509C6'),
            PublicKey('20EEEFCAE026EBB3C3C51E9AF86A97AA146B34A5463CFE468B37C3CB49682408')
        ])

    def test_can_use_bip32_derivation_with_passphrase(self):
        self._assert_bip32_child_public_keys('TREZOR', [
            PublicKey('3BE4759796DD507D6E410CD8C65121E7E42DAC69699A9058E1F7663A390122CE'),
            PublicKey('6B288C00800EC9FC0C30F35CEAFC2C5EC4066C2BE622822AAC70D67F215E5E6D'),
            PublicKey('1AC159878D327E578C0130767E960C265753CAD5215FC992F1F71C41D00EADA3')
        ])

    # endregion



from binascii import hexlify


class BufferReader:
    """Reads data from an in memory buffer."""

    def __init__(self, buffer, byte_order='little'):
        """Creates a reader with specified byte order."""
        self.buffer = buffer
        self.byte_order = byte_order
        self.offset = 0

    @property
    def eof(self):
        """Returns true if the reader is at eof."""
        return self.offset == len(self.buffer)

    def read_int(self, count):
        """Reads an integer."""
        return int.from_bytes(self.read_bytes(count), self.byte_order)

    def read_string(self, count):
        """Reads a string."""
        return self.read_bytes(count).decode('utf8')

    def read_hex_string(self, count):
        """Reads a hex string."""
        return hexlify(self.read_bytes(count)).decode('utf8').upper()

    def read_bytes(self, count):
        """Reads bytes."""
        buffer = self.buffer[self.offset:self.offset + count]
        self.offset += count
        return buffer



import unittest
from types import SimpleNamespace

from symbolchain.core.CryptoTypes import PublicKey
from symbolchain.core.TransactionDescriptorProcessor import TransactionDescriptorProcessor


class TransactionDescriptorProcessorTest(unittest.TestCase):
    # region lookup_value

    @staticmethod
    def _create_processor():
        transaction_descriptor = {
            'type': 'transfer',
            'timestamp': 12345,
            'signer': 'signer_name',
            'recipient': 'recipient_name',
            'message': 'hello world',
        }
        type_parsing_rules = {PublicKey: lambda name: name + ' PUBLICKEY'}
        processor = TransactionDescriptorProcessor(transaction_descriptor, type_parsing_rules)
        processor.set_type_hints({'signer': PublicKey, 'timestamp': int})
        return processor

    def test_cannot_lookup_value_when_descriptor_does_not_contain_key(self):
        # Arrange:
        processor = self._create_processor()

        # Act + Assert:
        with self.assertRaises(ValueError):
            processor.lookup_value('foo')

    def test_can_lookup_value_without_type_hint(self):
        # Arrange:
        processor = self._create_processor()

        # Act:
        message = processor.lookup_value('message')

        # Assert:
        self.assertEqual('hello world', message)

    def test_can_lookup_value_with_type_hint_but_without_custom_rule(self):
        # Arrange:
        processor = self._create_processor()

        # Act:
        timestamp = processor.lookup_value('timestamp')

        # Assert:
        self.assertEqual(12345, timestamp)

    def test_can_lookup_value_with_type_hint_and_with_custom_rule(self):
        # Arrange:
        processor = self._create_processor()

        # Act:
        signer = processor.lookup_value('signer')

        # Assert:
        self.assertEqual('signer_name PUBLICKEY', signer)

    # endregion

    # region copy_to

    def test_cannot_copy_to_when_descriptor_contains_fields_not_in_transaction(self):
        # Arrange:
        processor = self._create_processor()
        transaction = SimpleNamespace(type=None, timestamp=None, signer=None, recipient=None)  # missing message

        # Act + Assert:
        with self.assertRaises(ValueError):
            processor.copy_to(transaction)

    def test_can_copy_to_when_transaction_contains_exact_fields_in_descriptor(self):
        # Arrange:
        processor = self._create_processor()
        transaction = SimpleNamespace(type=None, timestamp=None, signer=None, recipient=None, message=None)

        # Act:
        processor.copy_to(transaction)

        # Assert:
        self.assertEqual('transfer', transaction.type)
        self.assertEqual(12345, transaction.timestamp)
        self.assertEqual('signer_name PUBLICKEY', transaction.signer)
        self.assertEqual('recipient_name', transaction.recipient)
        self.assertEqual('hello world', transaction.message)

    def test_can_copy_to_when_transaction_contains_fields_not_in_descriptor(self):
        # Arrange:
        processor = self._create_processor()
        transaction = SimpleNamespace(type=None, timestamp=None, signer=None, recipient=None, message=None, foo=None)

        # Act:
        processor.copy_to(transaction)

        # Assert:
        self.assertEqual('transfer', transaction.type)
        self.assertEqual(12345, transaction.timestamp)
        self.assertEqual('signer_name PUBLICKEY', transaction.signer)
        self.assertEqual('recipient_name', transaction.recipient)
        self.assertEqual('hello world', transaction.message)
        self.assertEqual(None, transaction.foo)

    def test_can_copy_to_when_ignore_keys_is_not_empty(self):
        # Arrange:
        processor = self._create_processor()
        transaction = SimpleNamespace(type=None, timestamp=None, signer=None, recipient=None, message=None)

        # Act:
        processor.copy_to(transaction, ['type', 'recipient'])

        # Assert:
        self.assertEqual(None, transaction.type)
        self.assertEqual(12345, transaction.timestamp)
        self.assertEqual('signer_name PUBLICKEY', transaction.signer)
        self.assertEqual(None, transaction.recipient)
        self.assertEqual('hello world', transaction.message)

    def test_can_copy_to_when_transaction_contains_iterable_attribute(self):
        # Arrange:
        transaction_descriptor = {
            'type': 'transfer',
            'signer': 'signer_name',
            'mosaics': [(1, 2), (3, 5)]
        }
        type_parsing_rules = {PublicKey: lambda name: name + ' PUBLICKEY'}
        processor = TransactionDescriptorProcessor(transaction_descriptor, type_parsing_rules)
        processor.set_type_hints({'signer': PublicKey})

        transaction = SimpleNamespace(type=None, signer=None, mosaics=[])

        # Act:
        processor.copy_to(transaction)

        # Assert:
        self.assertEqual('transfer', transaction.type)
        self.assertEqual('signer_name PUBLICKEY', transaction.signer)
        self.assertEqual([(1, 2), (3, 5)], transaction.mosaics)

    def test_can_copy_to_when_transaction_contains_tuple_attribute(self):
        # Arrange:
        transaction_descriptor = {
            'type': 'transfer',
            'signer': 'signer_name',
            'payload': (1, 2)
        }
        type_parsing_rules = {PublicKey: lambda name: name + ' PUBLICKEY'}
        processor = TransactionDescriptorProcessor(transaction_descriptor, type_parsing_rules)
        processor.set_type_hints({'signer': PublicKey})

        transaction = SimpleNamespace(type=None, signer=None, payload=None)

        # Act:
        processor.copy_to(transaction)

        # Assert:
        self.assertEqual('transfer', transaction.type)
        self.assertEqual('signer_name PUBLICKEY', transaction.signer)
        self.assertEqual((1, 2), transaction.payload)

    def test_can_copy_to_when_transaction_contains_bytes_attribute(self):
        # Arrange:
        transaction_descriptor = {
            'type': 'transfer',
            'signer': 'signer_name',
            'payload': b'FF554433'
        }
        type_parsing_rules = {PublicKey: lambda name: name + ' PUBLICKEY'}
        processor = TransactionDescriptorProcessor(transaction_descriptor, type_parsing_rules)
        processor.set_type_hints({'signer': PublicKey})

        transaction = SimpleNamespace(type=None, signer=None, payload=None)

        # Act:
        processor.copy_to(transaction)

        # Assert:
        self.assertEqual('transfer', transaction.type)
        self.assertEqual('signer_name PUBLICKEY', transaction.signer)
        self.assertEqual(b'FF554433', transaction.payload)

    # endregion

    # region set_type_hints

    def test_can_change_type_hints(self):
        # Arrange:
        processor = self._create_processor()
        transaction = SimpleNamespace(type=None, timestamp=None, signer=None, recipient=None, message=None)

        # Act:
        processor.set_type_hints({'recipient': PublicKey})
        processor.copy_to(transaction)

        # Assert:
        self.assertEqual('transfer', transaction.type)
        self.assertEqual(12345, transaction.timestamp)
        self.assertEqual('signer_name', transaction.signer)
        self.assertEqual('recipient_name PUBLICKEY', transaction.recipient)
        self.assertEqual('hello world', transaction.message)

    def test_can_clear_type_hints(self):
        # Arrange:
        processor = self._create_processor()
        transaction = SimpleNamespace(type=None, timestamp=None, signer=None, recipient=None, message=None)

        # Act:
        processor.set_type_hints(None)
        processor.copy_to(transaction)

        # Assert:
        self.assertEqual('transfer', transaction.type)
        self.assertEqual(12345, transaction.timestamp)
        self.assertEqual('signer_name', transaction.signer)
        self.assertEqual('recipient_name', transaction.recipient)
        self.assertEqual('hello world', transaction.message)

    # endregion



from abc import abstractmethod


class AddressTestDescriptor:
    def __init__(self, address_class, encoded_address, decoded_address):
        self.address_class = address_class
        self.encoded_address = encoded_address
        self.decoded_address = decoded_address


class BasicAddressTest:
    # pylint: disable=no-member

    def test_size_constant_is_correct(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act + Assert:
        self.assertEqual(len(test_descriptor.decoded_address), test_descriptor.address_class.SIZE)

    def test_can_create_from_address(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()
        original_address = test_descriptor.address_class(test_descriptor.encoded_address)

        # Act:
        address = test_descriptor.address_class(original_address)

        # Assert:
        self.assertEqual(test_descriptor.decoded_address, address.bytes)
        self.assertEqual(test_descriptor.encoded_address, str(address))

    def test_can_create_from_encoded_address(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act:
        address = test_descriptor.address_class(test_descriptor.encoded_address)

        # Assert:
        self.assertEqual(test_descriptor.decoded_address, address.bytes)
        self.assertEqual(test_descriptor.encoded_address, str(address))

    def test_can_create_from_decoded_address(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act:
        address = test_descriptor.address_class(test_descriptor.decoded_address)

        # Assert:
        self.assertEqual(test_descriptor.decoded_address, address.bytes)
        self.assertEqual(test_descriptor.encoded_address, str(address))

    @abstractmethod
    def get_test_descriptor(self):
        pass



from mnemonic import Mnemonic


class CodeWords:
    """Represents a group of words that encode some binary data."""

    def __init__(self, words):
        self.words = words
        self.separator = ' '

    def __eq__(self, other):
        return isinstance(other, CodeWords) and self.words == other.words

    def __str__(self):
        return self.separator.join(self.words)


class CodeWordsEncoder:
    """Encodes and decodes values using wordlists."""

    def __init__(self, wordlist='english'):
        self.wordlist = wordlist if not isinstance(wordlist, str) else Mnemonic(wordlist).wordlist
        self.wordlist_size = len(self.wordlist)

        if self.wordlist_size < 2:
            raise ValueError('wordlist must contain at least two words')

    def encode_int(self, value):
        """Encodes an integer into one or more code words."""
        seed = []
        while value:
            index = value % self.wordlist_size
            seed.append(self.wordlist[index])
            value = value // self.wordlist_size

        return CodeWords(seed if seed else [self.wordlist[0]])

    def decode_int(self, code_words):
        """Decodes one or more code words into an integer."""
        base = 1
        value = 0
        for word in code_words.words:
            index = self.wordlist.index(word)
            value += index * base
            base *= self.wordlist_size

        return value



from abc import abstractmethod

from symbolchain.core.CryptoTypes import PrivateKey, PublicKey, Signature

from .NemTestUtils import NemTestUtils


class KeyPairTestDescriptor:
    def __init__(self, key_pair_class, verifier_class, deterministic_private_key, expected_public_key):
        self.key_pair_class = key_pair_class
        self.verifier_class = verifier_class
        self.deterministic_private_key = deterministic_private_key
        self.expected_public_key = expected_public_key


class BasicKeyPairTest:
    # pylint: disable=no-member

    def test_can_create_key_pair_from_private_key(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        public_key = test_descriptor.expected_public_key
        private_key = test_descriptor.deterministic_private_key

        # Act:
        key_pair = test_descriptor.key_pair_class(private_key)

        # Assert:
        self.assertEqual(public_key, key_pair.public_key)
        self.assertEqual(private_key, key_pair.private_key)

    # region sign

    def test_sign_fills_the_signature(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        key_pair = test_descriptor.key_pair_class(PrivateKey.random())
        message = NemTestUtils.randbytes(21)

        # Act:
        signature = key_pair.sign(message)

        # Assert:
        self.assertNotEqual(bytes(), signature.bytes)

    def test_signatures_generated_for_same_data_by_same_key_pairs_are_equal(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        private_key = PrivateKey.random()
        key_pair1 = test_descriptor.key_pair_class(private_key)
        key_pair2 = test_descriptor.key_pair_class(private_key)
        message = NemTestUtils.randbytes(21)

        # Act:
        signature1 = key_pair1.sign(message)
        signature2 = key_pair2.sign(message)

        # Assert:
        self.assertEqual(signature1, signature2)

    def test_signatures_generated_for_same_data_by_different_key_pairs_are_different(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        key_pair1 = test_descriptor.key_pair_class(PrivateKey.random())
        key_pair2 = test_descriptor.key_pair_class(PrivateKey.random())
        message = NemTestUtils.randbytes(21)

        # Act:
        signature1 = key_pair1.sign(message)
        signature2 = key_pair2.sign(message)

        # Assert:
        self.assertNotEqual(signature1, signature2)

    # endregion

    # region verify

    def test_signature_can_be_verified(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        message = NemTestUtils.randbytes(21)
        key_pair = test_descriptor.key_pair_class(PrivateKey.random())
        signature = key_pair.sign(message)

        # Act:
        is_verified = test_descriptor.verifier_class(key_pair.public_key).verify(message, signature)

        # Assert:
        self.assertTrue(is_verified)

    def test_signature_cannot_be_verified_with_different_key_pair(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        message = NemTestUtils.randbytes(21)
        signature = test_descriptor.key_pair_class(PrivateKey.random()).sign(message)

        # Act:
        is_verified = test_descriptor.verifier_class(NemTestUtils.randcryptotype(PublicKey)).verify(message, signature)

        # Assert:
        self.assertFalse(is_verified)

    def test_signature_does_not_verify_when_message_modified(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        message = NemTestUtils.randbytes(21)
        key_pair = test_descriptor.key_pair_class(PrivateKey.random())
        signature = key_pair.sign(message)

        verifier = test_descriptor.verifier_class(key_pair.public_key)
        for i in range(0, 21):
            mutable_message = bytearray(message)
            mutable_message[i] ^= 0xFF

            # Act:
            is_verified = verifier.verify(bytes(mutable_message), signature)

            # Assert:
            self.assertFalse(is_verified)

    def test_signature_does_not_verify_when_signature_modified(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        message = NemTestUtils.randbytes(21)
        key_pair = test_descriptor.key_pair_class(PrivateKey.random())
        signature = key_pair.sign(message)

        verifier = test_descriptor.verifier_class(key_pair.public_key)
        for i in range(0, Signature.SIZE):
            mutable_signature = bytearray(signature.bytes)
            mutable_signature[i] ^= 0xFF

            # Act:
            is_verified = verifier.verify(message, Signature(bytes(mutable_signature)))

            # Assert:
            self.assertFalse(is_verified)

    # endregion

    @abstractmethod
    def get_test_descriptor(self):
        pass



import unittest
from binascii import unhexlify

from symbolchain.core.CryptoTypes import Hash256, PublicKey
from symbolchain.core.sym.Network import Address, Network
from symbolchain.tests.test.BasicAddressTest import AddressTestDescriptor, BasicAddressTest
from symbolchain.tests.test.BasicNetworkTest import BasicNetworkTest, NetworkTestDescriptor

PUBLIC_GENERATION_HASH_SEED = Hash256('57F7DA205008026C776CB6AED843393F04CD458E0AA2D9F1D5F31A402072B2D6')
PUBLIC_TEST_GENERATION_HASH_SEED = Hash256('45FBCF2F0EA36EFA7923C9BC923D6503169651F7FA4EFC46A8EAF5AE09057EBD')


class AddressTest(BasicAddressTest, unittest.TestCase):
    def get_test_descriptor(self):
        return AddressTestDescriptor(
            Address,
            'TBLYH55IHPS5QCCMNWR3GZWKV6WMCKPTNI7KSDA',
            unhexlify('985783F7A83BE5D8084C6DA3B366CAAFACC129F36A3EA90C'))


class NetworkTest(BasicNetworkTest, unittest.TestCase):
    def get_test_descriptor(self):
        return NetworkTestDescriptor(
            Network,
            PublicKey('C5FB65CB902623D93DF2E682FFB13F99D50FAC24D5FF2A42F68C7CA1772FE8A0'),
            [Address('NBLYH55IHPS5QCCMNWR3GZWKV6WMCKPTNKZIBEY'), Network.PUBLIC],
            [Address('TBLYH55IHPS5QCCMNWR3GZWKV6WMCKPTNI7KSDA'), Network.PUBLIC_TEST])

    def test_correct_predefined_networks_are_registered(self):
        self.assertEqual(4, len(Network.NETWORKS))
        self.assertEqual(['public', 'private', 'public_test', 'private_test'], [network.name for network in Network.NETWORKS])

        self._assert_network(Network.PUBLIC, 'public', 0x68)
        self.assertEqual(PUBLIC_GENERATION_HASH_SEED, Network.PUBLIC.generation_hash_seed)

        self._assert_network(Network.PRIVATE, 'private', 0x78)
        self.assertEqual(None, Network.PRIVATE.generation_hash_seed)

        self._assert_network(Network.PUBLIC_TEST, 'public_test', 0x98)
        self.assertEqual(PUBLIC_TEST_GENERATION_HASH_SEED, Network.PUBLIC_TEST.generation_hash_seed)

        self._assert_network(Network.PRIVATE_TEST, 'private_test', 0xA8)
        self.assertEqual(None, Network.PRIVATE_TEST.generation_hash_seed)



import sha3

NAMESPACE_FLAG = 1 << 63


def generate_mosaic_id(owner_address, nonce):
    """Generates a mosaic id from an owner address and a nonce."""
    hasher = sha3.sha3_256()
    hasher.update(nonce.to_bytes(4, 'little'))
    hasher.update(owner_address.bytes)
    digest = hasher.digest()

    result = int.from_bytes(digest[0:8], 'little')
    if result & NAMESPACE_FLAG:
        result -= NAMESPACE_FLAG

    return result


def generate_namespace_id(name, parent_namespace_id=0):
    """Generates a namespace id from a name and an optional parent namespace id."""
    hasher = sha3.sha3_256()
    hasher.update(parent_namespace_id.to_bytes(8, 'little'))
    hasher.update(name.encode('utf8'))
    digest = hasher.digest()

    result = int.from_bytes(digest[0:8], 'little')
    return result | NAMESPACE_FLAG


def generate_mosaic_alias_id(fully_qualified_name):
    """Generates a mosaic id from a fully qualified mosaic alias name."""
    return generate_namespace_path(fully_qualified_name)[-1]


def is_valid_namespace_name(name):
    """Returns true if a name is a valid namespace name."""
    def is_alphanum(character):
        return 'a' <= character <= 'z' or '0' <= character <= '9'

    return name and is_alphanum(name[0]) and all(is_alphanum(ch) or ch in ['_', '-'] for ch in name)


def generate_namespace_path(fully_qualified_name):
    """Parses a fully qualified namespace name into a path."""
    path = []
    parent_namespace_id = 0
    for name in fully_qualified_name.split('.'):
        if not is_valid_namespace_name(name):
            raise ValueError('fully qualified name is invalid due to invalid part name ({})'.format(fully_qualified_name))

        path.append(generate_namespace_id(name, parent_namespace_id))
        parent_namespace_id = path[-1]

    return path



import unittest

from symbolchain.core.Network import Network, NetworkLocator


class NetworkTest(unittest.TestCase):
    def test_equality_is_supported(self):
        # Arrange:
        foo_network = Network('foo', 0x55)

        # Act + Assert:
        self.assertEqual(foo_network, Network('foo', 0x55))
        self.assertNotEqual(foo_network, Network('Foo', 0x55))
        self.assertNotEqual(foo_network, Network('foo', 0x54))
        self.assertNotEqual(foo_network, None)

    def test_string_is_supported(self):
        self.assertEqual('foo', str(Network('foo', 0x55)))


PREDEFINED_NETWORKS = [Network('foo', 0x55), Network('bar', 0x37)]


class NetworkLocatorTest(unittest.TestCase):
    def test_can_find_well_known_network_by_name_single(self):
        self.assertEqual(PREDEFINED_NETWORKS[0], NetworkLocator.find_by_name(PREDEFINED_NETWORKS, 'foo'))
        self.assertEqual(PREDEFINED_NETWORKS[1], NetworkLocator.find_by_name(PREDEFINED_NETWORKS, 'bar'))

    def test_can_find_well_known_network_by_name_list(self):
        self.assertEqual(PREDEFINED_NETWORKS[0], NetworkLocator.find_by_name(PREDEFINED_NETWORKS, ['xxx', 'foo']))
        self.assertEqual(PREDEFINED_NETWORKS[1], NetworkLocator.find_by_name(PREDEFINED_NETWORKS, ['bar', 'yyy']))
        self.assertEqual(PREDEFINED_NETWORKS[0], NetworkLocator.find_by_name(PREDEFINED_NETWORKS, ['bar', 'foo']))

    def test_cannot_find_other_network_by_name(self):
        with self.assertRaises(StopIteration):
            NetworkLocator.find_by_name(PREDEFINED_NETWORKS, 'cat')

        with self.assertRaises(StopIteration):
            NetworkLocator.find_by_name(PREDEFINED_NETWORKS, ['cat', 'dog'])

    def test_can_find_well_known_network_by_identifier_single(self):
        self.assertEqual(PREDEFINED_NETWORKS[0], NetworkLocator.find_by_identifier(PREDEFINED_NETWORKS, 0x55))
        self.assertEqual(PREDEFINED_NETWORKS[1], NetworkLocator.find_by_identifier(PREDEFINED_NETWORKS, 0x37))

    def test_can_find_well_known_network_by_identifier_list(self):
        self.assertEqual(PREDEFINED_NETWORKS[0], NetworkLocator.find_by_identifier(PREDEFINED_NETWORKS, [0x88, 0x55]))
        self.assertEqual(PREDEFINED_NETWORKS[1], NetworkLocator.find_by_identifier(PREDEFINED_NETWORKS, [0x37, 0x99]))
        self.assertEqual(PREDEFINED_NETWORKS[0], NetworkLocator.find_by_identifier(PREDEFINED_NETWORKS, [0x37, 0x55]))

    def test_cannot_find_other_network_by_identifier(self):
        with self.assertRaises(StopIteration):
            NetworkLocator.find_by_identifier(PREDEFINED_NETWORKS, 0xFF)

        with self.assertRaises(StopIteration):
            NetworkLocator.find_by_identifier(PREDEFINED_NETWORKS, [0xFF, 0x88])



import unittest

from symbolchain.core.CryptoTypes import PrivateKey, PublicKey
from symbolchain.core.sym.KeyPair import KeyPair, Verifier
from symbolchain.tests.test.BasicKeyPairTest import BasicKeyPairTest, KeyPairTestDescriptor


class NetworkTest(BasicKeyPairTest, unittest.TestCase):
    def get_test_descriptor(self):
        deterministic_private_key = PrivateKey('E88283CE35FE74C89FFCB2D8BFA0A2CF6108BDC0D07606DEE34D161C30AC2F1E')
        expected_public_key = PublicKey('E29C5934F44482E7A9F50725C8681DE6CA63F49E5562DB7E5BC9EABA31356BAD')
        return KeyPairTestDescriptor(KeyPair, Verifier, deterministic_private_key, expected_public_key)



import os

import yaml

from ..CryptoTypes import PublicKey


class BatchOperations:
    """Provides facade-based batch operations."""

    class PrepareError(Exception):
        """Error that is thrown when a batch prepare operation fails."""

        def __init__(self, message):
            """Creates a new error."""
            Exception.__init__(self, message)

    def __init__(self, facade, output_file_prefix=''):
        """Creates batch operations around the specified facade."""
        self.facade = facade
        self.output_file_prefix = output_file_prefix

    def load_all(self, transactions_yaml_input):
        """Loads all transactions from YAML."""
        return [
            self.facade.transaction_factory.create(transaction_descriptor)
            for transaction_descriptor in yaml.load(transactions_yaml_input, Loader=yaml.SafeLoader)
        ]

    def _sign_one(self, transaction, private_key_storage, signature_storage, output_filename):
        # in symbol models, signer_public_key is bytes and needs to be wrapped
        signer_public_key = PublicKey(transaction.signer_public_key)
        signer_account_name = self.facade.account_descriptor_repository.find_by_public_key(signer_public_key).name
        signer_private_key = private_key_storage.load(signer_account_name)

        signature = self.facade.sign_transaction(self.facade.KeyPair(signer_private_key), transaction)
        transaction_hash = self.facade.hash_transaction(transaction)
        signature_storage.save(output_filename, transaction_hash, [signature])

    def sign_all(self, transactions, private_key_storage, signature_storage):
        """Signs multiple transactions and saves the signatures as multiple qrcodes."""
        for i, transaction in enumerate(transactions):
            self._sign_one(transaction, private_key_storage, signature_storage, 'sig_{}{}'.format(self.output_file_prefix, i))

    def prepare_all(self, transactions, signature_storage, output_directory):
        """Prepares multiple transactions by attaching signatures to each and producing files that can be sent to the network."""
        transaction_count = len(transactions)
        signature_groups = [signature_storage.load('sig_{}{}'.format(self.output_file_prefix, i)) for i in range(0, transaction_count)]

        for i, transaction in enumerate(transactions):
            (signed_transaction_hash, signatures) = signature_groups[i]

            transaction_hash = self.facade.hash_transaction(transaction)
            if signed_transaction_hash != transaction_hash:
                raise self.PrepareError('transaction hash at {} does not match signed transaction hash'.format(i))

            if not self.facade.verify_transaction(transaction, signatures[0]):
                raise self.PrepareError('transaction signature at {} does not verify'.format(i))

        for i, transaction in enumerate(transactions):
            (_, signatures) = signature_groups[i]

            prepared_transaction_buffer = self.facade.transaction_factory.attach_signature(transaction, signatures[0])
            with open(os.path.join(output_directory, 'payload_{}{}.dat'.format(self.output_file_prefix, i)), 'wb') as outfile:
                outfile.write(prepared_transaction_buffer)



import unittest
from binascii import unhexlify

from symbolchain.core.Bip32 import Bip32
from symbolchain.core.CryptoTypes import PrivateKey

DETERIMINISTIC_SEED = unhexlify('000102030405060708090A0B0C0D0E0F')
DETERIMINISTIC_MNEMONIC = 'cat swing flag economy stadium alone churn speed unique patch report train'


class Bip32Test(unittest.TestCase):
    # region from_seed

    def test_can_create_root_node(self):
        # Act:
        node = Bip32().from_seed(DETERIMINISTIC_SEED)

        # Assert:
        self._assert_bip32_node(
            node,
            unhexlify('90046A93DE5380A72B5E45010748567D5EA02BBF6522F979E05C0D8D8CA9FFFB'),
            unhexlify('2B4BE7F19EE27BBF30C667B642D5F4AA69FD169872F8FC3059C08EBAE2EB19E7'))

    def test_can_create_root_node_with_custom_curve_name(self):
        # Act:
        node = Bip32(curve_name='ed25519-keccak').from_seed(DETERIMINISTIC_SEED)

        # Assert:
        self._assert_bip32_node(
            node,
            unhexlify('9CFCA256458AAC0A0550A30DC7639D87364E4323BA61ED41454818E3317BAED0'),
            unhexlify('A3D76D92ACF784D68F4EA2F6DE5507A3520385237A80277132B6C8F3685601B2'))

    def test_can_derive_single_level_node(self):
        # Act:
        node = Bip32().from_seed(DETERIMINISTIC_SEED).derive_one(0)

        # Assert:
        self._assert_bip32_node(
            node,
            unhexlify('8B59AA11380B624E81507A27FEDDA59FEA6D0B779A778918A2FD3590E16E9C69'),
            unhexlify('68E0FE46DFB67E368C75379ACEC591DAD19DF3CDE26E63B93A8E704F1DADE7A3'))

    def _assert_well_known_children_from_deterministic_seed(self, child_node0, child_node1):
        self._assert_bip32_node(
            child_node0,
            unhexlify('B8E16D407C8837B46A9445C6417310F3C7A4DCD9B8FF2679C383E6DEF721AC11'),
            unhexlify('BB2724A538CFD64E4366FEB36BB982B954D58EA78F7163451B3B514EDD692159'))
        self._assert_bip32_node(
            child_node1,
            unhexlify('68CA2A058611AAC20CAFB4E1CCD70961E67D8C567390B3CBFC63D0E58BAE7153'),
            unhexlify('8C91D9F5D214A2E80A275E75A165F7022712F7AD52B7ECD45B3B6CC76154B571'))

    def test_can_derive_child_nodes_chained(self):
        # Act:
        node = Bip32().from_seed(DETERIMINISTIC_SEED)
        child_node0 = node.derive_one(44).derive_one(4343).derive_one(0).derive_one(0).derive_one(0)
        child_node1 = node.derive_one(44).derive_one(4343).derive_one(1).derive_one(0).derive_one(0)

        # Assert:
        self._assert_well_known_children_from_deterministic_seed(child_node0, child_node1)

    def test_can_derive_child_nodes_path(self):
        # Act:
        node = Bip32().from_seed(DETERIMINISTIC_SEED)
        child_node0 = node.derive_path([44, 4343, 0, 0, 0])
        child_node1 = node.derive_path([44, 4343, 1, 0, 0])

        # Assert:
        self._assert_well_known_children_from_deterministic_seed(child_node0, child_node1)

    def _assert_bip32_node(self, node, expected_chain_code, expected_private_key):
        self.assertEqual(expected_chain_code, node.chain_code)
        self.assertEqual(PrivateKey(expected_private_key), node.private_key)

    # endregion

    # region from_mnemonic

    def test_can_derive_child_nodes_from_mnemonic_with_password(self):
        # Act:
        node = Bip32().from_mnemonic(DETERIMINISTIC_MNEMONIC, 'TREZOR')
        child_node0 = node.derive_path([44, 4343, 0, 0, 0])
        child_node1 = node.derive_path([44, 4343, 1, 0, 0])
        child_node2 = node.derive_path([44, 4343, 2, 0, 0])

        # Assert:
        self.assertEqual(PrivateKey('1455FB18AB105444763EED593B7CA1C53EF6DDF8BDA1AB7004276FEAB1FCF222'), child_node0.private_key)
        self.assertEqual(PrivateKey('913967B3DFE1E94C50D5C92789DA194644D2A699E5BB75B171A3B68993B82A21'), child_node1.private_key)
        self.assertEqual(PrivateKey('AEC7C0143FC11F26FF5DB020492DACA7C8CF2640D2377AD3C721286472571602'), child_node2.private_key)

    # endregion



import yaml


class NodeDescriptor:
    """Represents a node."""

    def __init__(self, descriptor_yaml):
        """Creates a descriptor from a yaml container."""
        self.host = descriptor_yaml.get('host')
        self.roles = descriptor_yaml.get('roles') or []


class NodeDescriptorRepository:
    """Loads read-only node descriptors from YAML."""

    def __init__(self, yaml_input):
        """Loads node descriptors from the specified input."""
        descriptors_yaml = yaml_input if isinstance(yaml_input, list) else yaml.load(yaml_input, Loader=yaml.SafeLoader)
        self.descriptors = [NodeDescriptor(descriptor_yaml) for descriptor_yaml in descriptors_yaml]

    def find_all_by_role(self, role):
        """Finds all node descriptors with a matching role."""
        return [descriptor for descriptor in self.descriptors if not role or role in descriptor.roles]

    def find_all_not_by_role(self, role):
        """Finds all node descriptors without a matching role."""
        return [descriptor for descriptor in self.descriptors if role and role not in descriptor.roles]



import datetime
import unittest

from symbolchain.core.NetworkTimestamp import NetworkTimestamp, NetworkTimestampDatetimeConverter


# needed because NetworkTimestamp is abstract
class ConcreteNetworkTimestamp(NetworkTimestamp):
    def add_seconds(self, count):
        return ConcreteNetworkTimestamp(self.timestamp + 5 * count)


class NetworkTimestampTest(unittest.TestCase):
    def test_can_create_epochal_timestamp(self):
        # Act:
        timestamp = ConcreteNetworkTimestamp(0)

        # Assert:
        self.assertTrue(timestamp.epochal)
        self.assertEqual(0, timestamp.timestamp)

    def test_can_create_non_epochal_timestamp(self):
        # Act:
        timestamp = ConcreteNetworkTimestamp(123)

        # Assert:
        self.assertFalse(timestamp.epochal)
        self.assertEqual(123, timestamp.timestamp)

    def test_can_add_minutes(self):
        # Arrange:
        timestamp = ConcreteNetworkTimestamp(100)

        # Act:
        new_timestamp = timestamp.add_minutes(50)

        # Assert:
        self.assertEqual(100, timestamp.timestamp)
        self.assertEqual(100 + 60 * 5 * 50, new_timestamp.timestamp)

    def test_can_add_hours(self):
        # Arrange:
        timestamp = ConcreteNetworkTimestamp(100)

        # Act:
        new_timestamp = timestamp.add_hours(50)

        # Assert:
        self.assertEqual(100, timestamp.timestamp)
        self.assertEqual(100 + 60 * 60 * 5 * 50, new_timestamp.timestamp)

    def test_equality_is_supported(self):
        # Arrange:
        timestamp = ConcreteNetworkTimestamp(123)

        # Act + Assert:
        self.assertEqual(timestamp, ConcreteNetworkTimestamp(123))
        self.assertNotEqual(timestamp, ConcreteNetworkTimestamp(122))
        self.assertNotEqual(timestamp, ConcreteNetworkTimestamp(124))
        self.assertNotEqual(timestamp, None)

    def test_string_is_supported(self):
        self.assertEqual('123', str(ConcreteNetworkTimestamp(123)))


def create_converter():
    return NetworkTimestampDatetimeConverter(datetime.datetime(2020, 1, 2, 3), 'hours')


class NetworkTimestampDatetimeConverterTest(unittest.TestCase):
    def test_can_convert_epochal_timestamp_to_datetime(self):
        # Arrange:
        converter = create_converter()

        # Act:
        utc_timestamp = converter.to_datetime(0)

        # Assert:
        self.assertEqual(datetime.datetime(2020, 1, 2, 3), utc_timestamp)

    def test_can_convert_non_epochal_timestamp_to_datetime(self):
        # Arrange:
        converter = create_converter()

        # Act:
        utc_timestamp = converter.to_datetime(5)

        # Assert:
        self.assertEqual(datetime.datetime(2020, 1, 2, 3 + 5), utc_timestamp)

    def test_cannot_convert_datetime_before_epochal_timestamp(self):
        # Arrange:
        converter = create_converter()

        # Act + Assert::
        with self.assertRaises(ValueError):
            converter.to_difference(datetime.datetime(2020, 1, 2, 2))

    def test_can_convert_datetime_to_epochal_timestamp(self):
        # Arrange:
        converter = create_converter()

        # Act:
        raw_timestamp = converter.to_difference(datetime.datetime(2020, 1, 2, 3))

        # Assert:
        self.assertEqual(0, raw_timestamp)

    def test_can_convert_datetime_to_non_epochal_timestamp(self):
        # Arrange:
        converter = create_converter()

        # Act:
        raw_timestamp = converter.to_difference(datetime.datetime(2020, 1, 2, 3 + 5))

        # Assert:
        self.assertEqual(5, raw_timestamp)



import unittest
from binascii import unhexlify

from symbolchain.core.CryptoTypes import PublicKey
from symbolchain.core.nis1.Network import Address, Network
from symbolchain.tests.test.BasicAddressTest import AddressTestDescriptor, BasicAddressTest
from symbolchain.tests.test.BasicNetworkTest import BasicNetworkTest, NetworkTestDescriptor


class AddressTest(BasicAddressTest, unittest.TestCase):
    def get_test_descriptor(self):
        return AddressTestDescriptor(
            Address,
            'TCFGSLITSWMRROU2GO7FPMIUUDELUPSZUNUEZF33',
            unhexlify('988A692D13959918BA9A33BE57B114A0C8BA3E59A3684C977B'))


class NetworkTest(BasicNetworkTest, unittest.TestCase):
    def get_test_descriptor(self):
        return NetworkTestDescriptor(
            Network,
            PublicKey('D6C3845431236C5A5A907A9E45BD60DA0E12EFD350B970E7F58E3499E2E7A2F0'),
            [Address('NCFGSLITSWMRROU2GO7FPMIUUDELUPSZUNJABUMH'), Network.MAINNET],
            [Address('TCFGSLITSWMRROU2GO7FPMIUUDELUPSZUNUEZF33'), Network.TESTNET])

    def test_correct_predefined_networks_are_registered(self):
        self.assertEqual(2, len(Network.NETWORKS))
        self.assertEqual(['mainnet', 'testnet'], [network.name for network in Network.NETWORKS])

        self._assert_network(Network.MAINNET, 'mainnet', 0x68)
        self._assert_network(Network.TESTNET, 'testnet', 0x98)



import unittest

from symbolchain.core.sym.IdGenerator import (generate_mosaic_alias_id, generate_mosaic_id, generate_namespace_id, generate_namespace_path,
                                              is_valid_namespace_name)
from symbolchain.core.sym.Network import Address
from symbolchain.tests.test.NemTestUtils import NemTestUtils

TEST_VECTORS = {
    'uppercase': ['CAT.token', 'CAT.TOKEN', 'cat.TOKEN', 'cAt.ToKeN', 'CaT.tOkEn'],
    'improper_part': ['alpha.bet@.zeta', 'a!pha.beta.zeta', 'alpha.beta.ze^a'],
    'improper_qualified': ['.', '..', '...', '.a', 'b.', 'a..b', '.a.b', 'b.a.'],
}


class IdGeneratorTest(unittest.TestCase):
    # pylint: disable=too-many-public-methods

    # region generate_mosaic_id

    def test_generate_mosaic_id_generates_correct_id(self):
        # Act:
        mosaic_id = generate_mosaic_id(Address('TATNE7Q5BITMUTRRN6IB4I7FLSDRDWZA37JGO5Q'), 812613930)

        # Assert:
        self.assertEqual(0x570FB3ED9379624C, mosaic_id)

    def test_generate_mosaic_id_different_addresses_produce_different_ids(self):
        # Arrange:
        address1 = Address(NemTestUtils.randcryptotype(Address))
        address2 = Address(NemTestUtils.randcryptotype(Address))

        # Act:
        mosaic_id1 = generate_mosaic_id(address1, 812613930)
        mosaic_id2 = generate_mosaic_id(address2, 812613930)

        # Assert:
        self.assertNotEqual(mosaic_id1, mosaic_id2)

    def test_generate_mosaic_id_different_nonces_produce_different_ids(self):
        # Arrange:
        address = Address(NemTestUtils.randcryptotype(Address))

        # Act:
        mosaic_id1 = generate_mosaic_id(address, 812613930)
        mosaic_id2 = generate_mosaic_id(address, 812613931)

        # Assert:
        self.assertNotEqual(mosaic_id1, mosaic_id2)

    def test_generate_mosaic_id_has_high_bit_cleared(self):
        # Arrange:
        for _ in range(0, 1000):
            address = Address(NemTestUtils.randcryptotype(Address))

            # Act:
            mosaic_id = generate_mosaic_id(address, 812613930)

            # Assert:
            self.assertEqual(0, mosaic_id >> 63)

    # endregion

    # region generate_namespace_id

    def test_generate_namespace_id_generates_correct_root_id(self):
        # Act:
        namespace_id = generate_namespace_id('symbol')

        # Assert:
        self.assertEqual(0xA95F1F8A96159516, namespace_id)

    def test_generate_namespace_id_generates_correct_child_id(self):
        # Act:
        namespace_id = generate_namespace_id('xym', 0xA95F1F8A96159516)

        # Assert:
        self.assertEqual(0xE74B99BA41F4AFEE, namespace_id)

    def test_generate_namespace_id_different_names_produce_different_ids(self):
        # Act:
        namespace_id1 = generate_namespace_id('symbol')
        namespace_id2 = generate_namespace_id('Symbol')

        # Assert:
        self.assertNotEqual(namespace_id1, namespace_id2)

    def test_generate_namespace_id_different_parents_produce_different_ids(self):
        # Act:
        namespace_id1 = generate_namespace_id('symbol', 0xA95F1F8A96159516)
        namespace_id2 = generate_namespace_id('symbol', 0xA95F1F8A96159517)

        # Assert:
        self.assertNotEqual(namespace_id1, namespace_id2)

    def test_generate_namespace_id_has_high_bit_set(self):
        # Arrange:
        for i in range(0, 1000):
            # Act:
            namespace_id = generate_namespace_id('symbol', i)

            # Assert:
            self.assertEqual(1, namespace_id >> 63)

    # endregion

    # region generate_mosaic_alias_id

    def test_generate_mosaic_alias_id_generates_correct_id(self):
        # Act:
        mosaic_id = generate_mosaic_alias_id('cat.token')

        # Assert:
        self.assertEqual(0xA029E100621B2E33, mosaic_id)

    def test_generate_mosaic_alias_id_supports_multilevel_mosaics(self):
        # Act:
        mosaic_id = generate_mosaic_alias_id('foo.bar.baz.xyz')

        # Assert:
        namespace_id = generate_namespace_id('baz', generate_namespace_id('bar', generate_namespace_id('foo')))
        expected_mosaic_id = generate_namespace_id('xyz', namespace_id)
        self.assertEqual(expected_mosaic_id, mosaic_id)

    def _assert_rejected_by_generate_mosaic_alias_id(self, names):
        for name in names:
            with self.assertRaises(ValueError):
                generate_mosaic_alias_id(name)

    def test_generate_mosaic_alias_id_rejects_uppercase_characters(self):
        self._assert_rejected_by_generate_mosaic_alias_id(TEST_VECTORS['uppercase'])

    def test_generate_mosaic_alias_id_rejects_improper_part_names(self):
        self._assert_rejected_by_generate_mosaic_alias_id(TEST_VECTORS['improper_part'])

    def test_generate_mosaic_alias_id_rejects_improper_qualified_names(self):
        self._assert_rejected_by_generate_mosaic_alias_id(TEST_VECTORS['improper_qualified'])

    def test_generate_mosaic_alias_id_rejects_empty_string(self):
        self._assert_rejected_by_generate_mosaic_alias_id([''])

    # endregion

    # region generate_namespace_path

    def test_generate_namespace_path_generates_correct_root_id(self):
        # Act:
        path = generate_namespace_path('cat')

        # Assert:
        self.assertEqual([0xB1497F5FBA651B4F], path)

    def test_generate_namespace_path_generates_correct_child_id(self):
        # Act:
        path = generate_namespace_path('cat.token')

        # Assert:
        self.assertEqual([0xB1497F5FBA651B4F, 0xA029E100621B2E33], path)

    def test_generate_namespace_path_supports_multilevel_namespaces(self):
        # Act:
        path = generate_namespace_path('foo.bar.baz.xyz')

        # Assert:
        expected_path = [generate_namespace_id('foo')]
        for name in ['bar', 'baz', 'xyz']:
            expected_path.append(generate_namespace_id(name, expected_path[-1]))

        self.assertEqual(expected_path, path)

    def _assert_rejected_by_generate_namespace_path(self, names):
        for name in names:
            with self.assertRaises(ValueError):
                generate_namespace_path(name)

    def test_generate_namespace_path_rejects_uppercase_characters(self):
        self._assert_rejected_by_generate_namespace_path(TEST_VECTORS['uppercase'])

    def test_generate_namespace_path_rejects_improper_part_names(self):
        self._assert_rejected_by_generate_namespace_path(TEST_VECTORS['improper_part'])

    def test_generate_namespace_path_rejects_improper_qualified_names(self):
        self._assert_rejected_by_generate_namespace_path(TEST_VECTORS['improper_qualified'])

    def test_generate_namespace_path_rejects_empty_string(self):
        self._assert_rejected_by_generate_namespace_path([''])

    # endregion

    # region is_valid_namespace_name

    def test_is_valid_namespace_name_returns_true_when_all_characters_are_alphanumeric(self):
        for name in ['a', 'be', 'cat', 'doom', '09az09', 'az09az']:
            self.assertTrue(is_valid_namespace_name(name))

    def test_is_valid_namespace_name_returns_true_when_name_contains_separator(self):
        for name in ['al-ce', 'al_ce', 'alice-', 'alice_']:
            self.assertTrue(is_valid_namespace_name(name))

    def test_is_valid_namespace_name_returns_false_when_name_starts_with_separator(self):
        for name in ['-alice', '_alice']:
            self.assertFalse(is_valid_namespace_name(name))

    def test_is_valid_namespace_name_returns_false_when_any_character_is_invalid(self):
        for name in ['al.ce', 'alIce', 'al ce', 'al@ce', 'al#ce']:
            self.assertFalse(is_valid_namespace_name(name))

    # endregion



import secrets

from .ByteArray import ByteArray


class Hash256(ByteArray):
    """Represents a 256-bit hash."""

    SIZE = 32

    def __init__(self, hash256):
        """Creates a hash from bytes or a hex string."""
        super().__init__(self.SIZE, hash256, Hash256)

    @staticmethod
    def zero():
        """Creates a zeroed hash."""
        return Hash256(bytes([0] * Hash256.SIZE))


class PrivateKey(ByteArray):
    """Represents a private key."""

    SIZE = 32

    def __init__(self, private_key):
        """Creates a private key from bytes or a hex string."""
        super().__init__(self.SIZE, private_key, PrivateKey)

    @staticmethod
    def random():
        """Generates a random private key."""
        return PrivateKey(secrets.token_bytes(PrivateKey.SIZE))


class PublicKey(ByteArray):
    """Represents a public key."""

    SIZE = 32

    def __init__(self, public_key):
        """Creates a public key from bytes or a hex string."""
        super().__init__(self.SIZE, public_key.bytes if isinstance(public_key, PublicKey) else public_key, PublicKey)


class Signature(ByteArray):
    """Represents a signature."""

    SIZE = 64

    def __init__(self, signature):
        """Creates a signature from bytes or a hex string."""
        super().__init__(self.SIZE, signature, Signature)

    @staticmethod
    def zero():
        """Creates a zeroed signature."""
        return Signature(bytes([0] * Signature.SIZE))



import unittest

from symbolchain.core.CryptoTypes import PrivateKey, PublicKey
from symbolchain.core.nis1.KeyPair import KeyPair, Verifier
from symbolchain.tests.test.BasicKeyPairTest import BasicKeyPairTest, KeyPairTestDescriptor


class NetworkTest(BasicKeyPairTest, unittest.TestCase):
    def get_test_descriptor(self):
        deterministic_private_key = PrivateKey('ED4C70D78104EB11BCD73EBDC512FEBC8FBCEB36A370C957FF7E266230BB5D57')  # reversed
        expected_public_key = PublicKey('D6C3845431236C5A5A907A9E45BD60DA0E12EFD350B970E7F58E3499E2E7A2F0')
        return KeyPairTestDescriptor(KeyPair, Verifier, deterministic_private_key, expected_public_key)



import importlib
from binascii import hexlify

from symbol_catbuffer.EmbeddedTransactionBuilderFactory import EmbeddedTransactionBuilderFactory
from symbol_catbuffer.NetworkTypeDto import NetworkTypeDto
from symbol_catbuffer.TransactionBuilderFactory import TransactionBuilderFactory

from ..CryptoTypes import PublicKey
from ..TransactionDescriptorProcessor import TransactionDescriptorProcessor
from .IdGenerator import generate_mosaic_id, generate_namespace_id
from .Network import Address


def get_catbuffer_class(name):
    return getattr(importlib.import_module('symbol_catbuffer.{}'.format(name)), name)


def name_to_enum_value(mapping, enum_type, enum_value_name):
    if enum_value_name not in mapping:
        raise RuntimeError('unknown value {} for type {}'.format(enum_value_name, enum_type))
    return mapping[enum_value_name]


def create_flags_parser(name):
    string_to_enum = dict(map(lambda key: (key.name.lower(), key), get_catbuffer_class(name)))

    def parser(flags):
        if not isinstance(flags, str):
            return flags

        enum_array = list(map(lambda flag_name: name_to_enum_value(string_to_enum, name, flag_name), flags.split(' ')))
        return enum_array

    return parser


def create_array_parser(name):
    string_to_enum = dict(map(lambda key: (key.name.lower(), key), get_catbuffer_class(name)))

    def parser(flags):
        if not isinstance(flags, list) or not isinstance(flags[0], str):
            return flags

        enum_array = list(map(lambda flag_name: name_to_enum_value(string_to_enum, name, flag_name), flags))
        return enum_array

    return parser


def create_enum_parser(name):
    string_to_enum = dict(map(lambda key: (key.name.lower(), key), get_catbuffer_class(name)))

    def parser(enum_value):
        if not isinstance(enum_value, str):
            return enum_value

        return name_to_enum_value(string_to_enum, name, enum_value)

    return parser


class TransactionFactory:
    """Factory for creating transactions."""

    def __init__(self, network, type_parsing_rules=None):
        """Creates a factory for the specified network."""
        self.network = network
        self.type_parsing_rules = type_parsing_rules

    def _extend_type_parsing_rules(self):
        default_rules = {
            'MosaicFlagsDto': create_flags_parser('MosaicFlagsDto'),
            'AccountRestrictionFlagsDto': create_flags_parser('AccountRestrictionFlagsDto'),

            'array_EntityTypeDto': create_array_parser('EntityTypeDto'),

            'AliasActionDto': create_enum_parser('AliasActionDto'),
            'LinkActionDto': create_enum_parser('LinkActionDto'),
            'LockHashAlgorithmDto': create_enum_parser('LockHashAlgorithmDto'),
            'MosaicRestrictionTypeDto': create_enum_parser('MosaicRestrictionTypeDto'),
            'MosaicSupplyChangeActionDto': create_enum_parser('MosaicSupplyChangeActionDto'),
            'NamespaceRegistrationTypeDto': create_enum_parser('NamespaceRegistrationTypeDto')
        }

        if not self.type_parsing_rules:
            return default_rules

        return {**default_rules, **self.type_parsing_rules}

    def _create(self, transaction_descriptor, factory_class):
        processor = TransactionDescriptorProcessor(transaction_descriptor, self._extend_type_parsing_rules())

        processor.set_type_hints({'signer_public_key': PublicKey})

        transaction_type = processor.lookup_value('type')
        transaction = factory_class.create_by_name(
            transaction_type,
            processor.lookup_value('signer_public_key'),
            NetworkTypeDto(self.network.identifier))

        all_type_hints = self._build_type_hints_map(transaction)
        processor.set_type_hints(all_type_hints)
        processor.copy_to(transaction, ['type', 'signer_public_key'])

        # autogenerate artifact ids
        if 'namespaceRegistration' == transaction_type:
            transaction.id = generate_namespace_id(transaction.name, transaction.parent_id)
        elif 'mosaicDefinition' == transaction_type:
            address = self.network.public_key_to_address(PublicKey(transaction.signer_public_key))
            transaction.id = generate_mosaic_id(address, transaction.nonce)

        # auto encode strings
        for key in transaction_descriptor.keys():
            if key in all_type_hints:
                continue

            value = getattr(transaction, key)
            if isinstance(value, str):
                setattr(transaction, key, value.encode('utf8'))

        return transaction

    def create(self, transaction_descriptor):
        """Creates a transaction from a transaction descriptor."""
        return self._create(transaction_descriptor, TransactionBuilderFactory)

    def create_embedded(self, transaction_descriptor):
        """Creates an embedded transaction from a transaction descriptor."""
        return self._create(transaction_descriptor, EmbeddedTransactionBuilderFactory)

    @staticmethod
    def attach_signature(transaction, signature):
        """Attaches a signature to a transaction."""
        transaction.signature = signature.bytes

        transaction_buffer = transaction.serialize()
        json_payload = '{{"payload": "{}"}}'.format(hexlify(transaction_buffer).decode('utf8').upper())
        return json_payload.encode('utf8')

    @staticmethod
    def _build_type_hints_map(transaction):
        mapping = {
            'UnresolvedAddressDto': Address,
            'KeyDto': PublicKey
        }

        type_hints = {}
        for key, hint in transaction.type_hints.items():
            rule_name = None
            if hint.startswith('array[enum:'):
                rule_name = 'array_' + hint[len('array[enum:'):-1]
            elif hint.startswith('enum:') and hint not in ('enum:NetworkTypeDto', 'enum:EntityTypeDto'):
                rule_name = hint[len('enum:'):]
            elif hint.endswith('FlagsDto'):
                rule_name = hint
            else:
                rule_name = mapping.get(hint, None)

            if rule_name:
                type_hints[key] = rule_name

        return type_hints



import math
import random
import unittest
from binascii import unhexlify

from mnemonic import Mnemonic

from symbolchain.core.DiceMnemonicGenerator import DiceMnemonicGenerator


def mnemonic_to_seed(words):
    return Mnemonic('english').to_entropy(words)


class DiceMnemonicGeneratorTest(unittest.TestCase):
    # region add_roll

    def test_can_add_single_roll_in_range(self):
        # Arrange:
        for roll in [1, 2, 3, 4, 5, 6]:
            generator = DiceMnemonicGenerator()

            # Act:
            generator.add_roll(roll)

            # Assert:
            self.assertEqual([roll], generator.rolls)

    def test_can_add_multiple_rolls_in_range(self):
        # Arrange:
        generator = DiceMnemonicGenerator()

        # Act:
        for roll in [1, 6, 3, 5, 4, 2]:
            generator.add_roll(roll)

        # Assert:
        self.assertEqual([1, 6, 3, 5, 4, 2], generator.rolls)

    def test_cannot_add_roll_out_of_range(self):
        # Arrange:
        for roll in [-1, 0, 7, 8, 100]:
            generator = DiceMnemonicGenerator()

            # Act + Assert:
            with self.assertRaises(ValueError):
                generator.add_roll(roll)

    # endregion

    # region frequencies

    def test_frequencies_are_initially_zero(self):
        # Arrange:
        generator = DiceMnemonicGenerator()

        # Act:
        frequencies = generator.frequencies()

        # Assert:
        self.assertEqual([0] * 6, frequencies)

    def test_frequencies_are_calculated_correctly(self):
        # Arrange:
        generator = DiceMnemonicGenerator()

        rolls = [1] * 17 + [2] * 15 + [3] * 9 + [4] * 12 + [5] * 10 + [6] * 19
        random.shuffle(rolls)
        for roll in rolls:
            generator.add_roll(roll)

        # Act:
        frequencies = generator.frequencies()

        # Assert:
        self.assertEqual([17, 15, 9, 12, 10, 19], frequencies)

    def test_frequencies_are_calculated_correctly_custom_die(self):
        # Arrange:
        generator = DiceMnemonicGenerator(num_die_sides=8)

        rolls = [1] * 17 + [2] * 15 + [3] * 9 + [4] * 12 + [5] * 10 + [6] * 19 + [7] * 8 + [8] * 14
        random.shuffle(rolls)
        for roll in rolls:
            generator.add_roll(roll)

        # Act:
        frequencies = generator.frequencies()

        # Assert:
        self.assertEqual([17, 15, 9, 12, 10, 19, 8, 14], frequencies)

    # endregion

    # region to_mnemonic (shrink_wrap=*)

    def _assert_mnemonic_can_be_created_with_max_entropy(self, shrink_wrap):
        # Arrange:
        generator = DiceMnemonicGenerator()

        for roll in [1, 2] * 50:
            generator.add_roll(roll)

        # Act:
        (mnemonic, entropy_bits) = generator.to_mnemonic(shrink_wrap)
        seed = mnemonic_to_seed(mnemonic)

        # Assert:
        self.assertEqual(24, len(mnemonic.split()))
        self.assertEqual(256, entropy_bits)
        self.assertEqual(unhexlify('C6DDDB140A8D03482E696D47A911E89634049EFA7B2AF3297105FFEAE4EADB7C'), seed)

    def _assert_mnemonic_caps_entropy_at_256bits(self, shrink_wrap):
        # Arrange:
        generator = DiceMnemonicGenerator()

        for roll in [1, 2] * 100:
            generator.add_roll(roll)

        # Act:
        (mnemonic, entropy_bits) = generator.to_mnemonic(shrink_wrap)
        seed = mnemonic_to_seed(mnemonic)

        # Assert:
        self.assertEqual(24, len(mnemonic.split()))
        self.assertEqual(256, entropy_bits)
        self.assertEqual(unhexlify('DD979CB2857F8D1F48837A53A8DF9A2FA9FEF5D4CBC563C252EC9815267E7437'), seed)

    # endregion

    # region to_mnemonic (shrink_wrap=False)

    def test_mnemonic_returns_correct_mnemonic_for_known_rolls(self):
        # Arrange:
        generator = DiceMnemonicGenerator()

        for roll in [1, 2, 3, 4, 5, 6]:
            generator.add_roll(roll)

        # Act:
        (mnemonic, entropy_bits) = generator.to_mnemonic()
        seed = mnemonic_to_seed(mnemonic)

        # Assert:
        self.assertEqual(24, len(mnemonic.split()))
        self.assertEqual(-math.log2(1/6) * 6, entropy_bits)
        self.assertEqual(unhexlify('8D969EEF6ECAD3C29A3A629280E686CF0C3F5D5A86AFF3CA12020C923ADC6C92'), seed)

    def test_mnemonic_returns_correct_mnemonic_for_known_rolls_custom_die(self):
        # Arrange:
        generator = DiceMnemonicGenerator(num_die_sides=8)

        for roll in [1, 2, 3, 4, 5, 6, 7, 8]:
            generator.add_roll(roll)

        # Act:
        (mnemonic, entropy_bits) = generator.to_mnemonic()
        seed = mnemonic_to_seed(mnemonic)

        # Assert:
        self.assertEqual(24, len(mnemonic.split()))
        self.assertEqual(-math.log2(1/8) * 8, entropy_bits)
        self.assertEqual(unhexlify('EF797C8118F02DFB649607DD5D3F8C7623048C9C063D532CC95C5ED7A898A64F'), seed)

    def test_mnemonic_has_no_min_entropy(self):
        # Arrange:
        generator = DiceMnemonicGenerator()

        # Act:
        (mnemonic, entropy_bits) = generator.to_mnemonic()
        seed = mnemonic_to_seed(mnemonic)

        # Assert:
        self.assertEqual(24, len(mnemonic.split()))
        self.assertEqual(0, entropy_bits)
        self.assertEqual(unhexlify('E3B0C44298FC1C149AFBF4C8996FB92427AE41E4649B934CA495991B7852B855'), seed)

    def test_mnemonic_can_be_created_with_max_entropy(self):
        self._assert_mnemonic_can_be_created_with_max_entropy(False)

    def test_mnemonic_caps_entropy_at_256bits(self):
        self._assert_mnemonic_caps_entropy_at_256bits(False)

    # endregion

    # region to_mnemonic (shrink_wrap=True)

    def test_mnemonic_shrink_wrap_returns_correct_mnemonic_for_known_rolls(self):
        # Arrange:
        generator = DiceMnemonicGenerator()

        for roll in [2, 3, 4, 2, 6, 5, 2, 6] * 6 + [3, 3]:
            generator.add_roll(roll)

        # Act:
        (mnemonic, entropy_bits) = generator.to_mnemonic(shrink_wrap=True)
        seed = mnemonic_to_seed(mnemonic)

        # Assert:
        self.assertEqual(12, len(mnemonic.split()))
        self.assertEqual(128, entropy_bits)
        self.assertEqual(unhexlify('49DBC749E6024695F3580B722E148341'), seed)

    def test_mnemonic_shrink_wrap_returns_correct_mnemonic_for_known_rolls_custom_die(self):
        # Arrange:
        generator = DiceMnemonicGenerator(num_die_sides=8)

        for roll in [2, 3, 4, 2, 6, 5, 2, 6] * 5 + [8, 7, 6]:
            generator.add_roll(roll)

        # Act:
        (mnemonic, entropy_bits) = generator.to_mnemonic(shrink_wrap=True)
        seed = mnemonic_to_seed(mnemonic)

        # Assert:
        self.assertEqual(12, len(mnemonic.split()))
        self.assertEqual(128, entropy_bits)
        self.assertEqual(unhexlify('3CCAD1E97E6770F667DD982F7D1408A4'), seed)

    def test_mnemonic_shrink_wrap_fails_with_entropy_less_than_128bits(self):
        # Arrange:
        for roll_count in [0, 25, 49]:
            generator = DiceMnemonicGenerator()

            for roll in [1] * roll_count:
                generator.add_roll(roll)

            with self.assertRaises(ValueError):
                generator.to_mnemonic(shrink_wrap=True)

    def test_mnemonic_shrink_wrap_can_be_created_with_min_entropy(self):
        # Arrange:
        generator = DiceMnemonicGenerator()

        for roll in [1, 2] * 25:
            generator.add_roll(roll)

        # Act:
        (mnemonic, entropy_bits) = generator.to_mnemonic(shrink_wrap=True)
        seed = mnemonic_to_seed(mnemonic)

        # Assert:
        self.assertEqual(12, len(mnemonic.split()))
        self.assertEqual(128, entropy_bits)
        self.assertEqual(unhexlify('11AB281EDD8D8F9789EC130530308357'), seed)

    def test_mnemonic_shrink_wrap_truncates_entropy_to_mnemonic_word_multiple(self):
        # Arrange:
        generator = DiceMnemonicGenerator()

        for roll in [1, 2] * 49 + [1]:
            generator.add_roll(roll)

        # Act:
        (mnemonic, entropy_bits) = generator.to_mnemonic(shrink_wrap=True)
        seed = mnemonic_to_seed(mnemonic)

        # Assert:
        self.assertEqual(21, len(mnemonic.split()))
        self.assertEqual(224, entropy_bits)
        self.assertEqual(unhexlify('FFC3D4A72768FB950D1D58BE65D509749416DE4A7C7C3C309382CBC7'), seed)

    def test_mnemonic_shrink_wrap_can_be_created_with_max_entropy(self):
        self._assert_mnemonic_can_be_created_with_max_entropy(True)

    def test_mnemonic_shrink_wrap_caps_entropy_at_256bits(self):
        self._assert_mnemonic_caps_entropy_at_256bits(True)

    # endregion



import unittest
from binascii import hexlify

from symbolchain.core.ByteArray import ByteArray
from symbolchain.tests.test.NemTestUtils import NemTestUtils

FIXED_SIZE = 24
TEST_BYTES = bytes([
    0xC5, 0xFB, 0x65, 0xCB, 0x90, 0x26, 0x23, 0xD9,
    0x3D, 0xF2, 0xE6, 0x82, 0xFF, 0xB1, 0x3F, 0x99,
    0xD5, 0x0F, 0xAC, 0x24, 0xD5, 0xFF, 0x2A, 0x42
])
TEST_HEX = 'C5FB65CB902623D93DF2E682FFB13F99D50FAC24D5FF2A42'


def random_hex_string(size):
    return hexlify(NemTestUtils.randbytes(size)).decode('utf8')


class ByteArrayTest(unittest.TestCase):
    def test_can_create_byte_array_with_correct_number_of_bytes(self):
        # Act:
        byte_array = ByteArray(FIXED_SIZE, TEST_BYTES)

        # Assert:
        self.assertEqual(TEST_BYTES, byte_array.bytes)

    def test_cannot_create_byte_array_with_incorrect_number_of_bytes(self):
        for size in [0, FIXED_SIZE - 1, FIXED_SIZE + 1]:
            with self.assertRaises(ValueError):
                ByteArray(FIXED_SIZE, NemTestUtils.randbytes(size))

    def test_can_create_byte_array_with_correct_number_of_hex_characters(self):
        # Act:
        byte_array = ByteArray(FIXED_SIZE, TEST_HEX)

        # Assert:
        self.assertEqual(TEST_BYTES, byte_array.bytes)

    def test_cannot_create_byte_array_with_incorrect_number_of_hex_characters(self):
        for size in [0, FIXED_SIZE - 1, FIXED_SIZE + 1]:
            with self.assertRaises(ValueError):
                ByteArray(FIXED_SIZE, random_hex_string(size))

    def test_equality_is_supported(self):
        # Arrange:
        byte_array = ByteArray(FIXED_SIZE, TEST_BYTES)
        byte_array_with_explicit_tag = ByteArray(FIXED_SIZE, TEST_BYTES, str)

        # Act + Assert:
        self.assertEqual(byte_array, ByteArray(FIXED_SIZE, TEST_BYTES))
        self.assertEqual(byte_array, ByteArray(FIXED_SIZE, TEST_HEX))
        self.assertEqual(byte_array_with_explicit_tag, ByteArray(FIXED_SIZE, TEST_BYTES, str))
        self.assertEqual(byte_array_with_explicit_tag, ByteArray(FIXED_SIZE, TEST_HEX, str))

        self.assertNotEqual(byte_array, ByteArray(FIXED_SIZE, NemTestUtils.randbytes(FIXED_SIZE)))
        self.assertNotEqual(byte_array, ByteArray(FIXED_SIZE, random_hex_string(FIXED_SIZE)))
        self.assertNotEqual(byte_array, byte_array_with_explicit_tag)
        self.assertNotEqual(byte_array, None)

    def test_string_is_supported(self):
        self.assertEqual(TEST_HEX, str(ByteArray(FIXED_SIZE, TEST_BYTES)))



from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519

from ..CryptoTypes import PrivateKey, PublicKey, Signature


class KeyPair:
    """Represents an ED25519 private and public key."""

    def __init__(self, private_key):
        """Creates a key pair from a private key."""
        self._sk = ed25519.Ed25519PrivateKey.from_private_bytes(private_key.bytes)

    @property
    def public_key(self):
        """Gets the public key."""
        return PublicKey(self._sk.public_key().public_bytes(serialization.Encoding.Raw, serialization.PublicFormat.Raw))

    @property
    def private_key(self):
        """Gets the private key."""
        return PrivateKey(self._sk.private_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PrivateFormat.Raw,
                encryption_algorithm=serialization.NoEncryption()))

    def sign(self, message):
        """Signs a message with the private key."""
        return Signature(self._sk.sign(message))


class Verifier:
    """Verifies signatures signed by a single key pair."""

    def __init__(self, public_key):
        """Creates a verifier from a public key."""
        self._pk = ed25519.Ed25519PublicKey.from_public_bytes(public_key.bytes)

    def verify(self, message, signature):
        """Verifies a message signature."""
        try:
            self._pk.verify(signature.bytes, message)
            return True
        except InvalidSignature:
            return False



import sha3

from ..CryptoTypes import Hash256, PrivateKey, PublicKey
from ..Network import NetworkLocator
from ..nis1.KeyPair import KeyPair, Verifier
from ..nis1.Network import Address, Network
from ..nis1.TransactionFactory import TransactionFactory


class NisFacade:
    """Facade used to interact with nis blockchain."""

    BIP32_COIN_ID = 43
    BIP32_CURVE_NAME = 'ed25519-keccak'

    KeyPair = KeyPair
    Verifier = Verifier

    def __init__(self, nis_network_name, account_descriptor_repository=None):
        """Creates a nis facade."""
        self.network = NetworkLocator.find_by_name(Network.NETWORKS, nis_network_name)
        self.account_descriptor_repository = account_descriptor_repository
        self.transaction_factory = self._create_nis_transaction_factory()

    def _create_nis_transaction_factory(self):
        type_parsing_rules = None
        if self.account_descriptor_repository:
            type_parsing_rules = self.account_descriptor_repository.to_type_parsing_rules_map(self._create_nis_type_to_property_mapping())

        return TransactionFactory(self.network, type_parsing_rules)

    @staticmethod
    def _create_nis_type_to_property_mapping():
        return {
            Address: 'address',
            PublicKey: 'public_key'
        }

    @staticmethod
    def hash_transaction(transaction):
        """Hashes a nis transaction."""
        return Hash256(sha3.keccak_256(transaction.serialize()).digest())

    @staticmethod
    def sign_transaction(key_pair, transaction):
        """Signs a nis transaction."""
        return key_pair.sign(transaction.serialize())

    @staticmethod
    def verify_transaction(transaction, signature):
        """Verifies a nis transaction."""
        return Verifier(transaction.signer_public_key).verify(transaction.serialize(), signature)

    @staticmethod
    def bip32_node_to_key_pair(bip32_node):
        """Derives a nis KeyPair from a BIP32 node."""
        # BIP32 private keys should be used as is, so reverse here to counteract reverse in KeyPair
        return KeyPair(PrivateKey(bip32_node.private_key.bytes[::-1]))



import yaml

from .CryptoTypes import PublicKey


class AccountDescriptor:
    """Represents an account."""

    def __init__(self, descriptor_yaml):
        """Creates a descriptor from a yaml container."""
        self.public_key = descriptor_yaml.get('public_key')
        if self.public_key:
            self.public_key = PublicKey(self.public_key)

        self.address = descriptor_yaml.get('address')
        self.name = descriptor_yaml.get('name')
        self.roles = descriptor_yaml.get('roles') or []


class AccountDescriptorRepository:
    """Loads read-only account descriptors from YAML."""

    def __init__(self, yaml_input):
        """Loads account descriptors from the specified input."""
        descriptors_yaml = yaml_input if isinstance(yaml_input, list) else yaml.load(yaml_input, Loader=yaml.SafeLoader)
        self.descriptors = [AccountDescriptor(descriptor_yaml) for descriptor_yaml in descriptors_yaml]

    def try_find_by_name(self, name):
        """Finds the account descriptor with a matching name or None if no matching descriptors are found."""
        return next((descriptor for descriptor in self.descriptors if name == descriptor.name), None)

    def find_by_public_key(self, public_key):
        """Finds the account descriptor with a matching public key."""
        return next(descriptor for descriptor in self.descriptors if descriptor.public_key and public_key == descriptor.public_key)

    def find_all_by_role(self, role):
        """Finds all account descriptors with a matching role."""
        return [descriptor for descriptor in self.descriptors if not role or role in descriptor.roles]

    def _lookup_account_descriptor_field(self, value, property_name, target_class):
        account_descriptor = self.try_find_by_name(value)
        if account_descriptor and hasattr(account_descriptor, property_name):
            return target_class(getattr(account_descriptor, property_name))

        return target_class(value)

    def _bind_lookup_account_descriptor_field(self, property_name, target_class):
        return lambda value: self._lookup_account_descriptor_field(value, property_name, target_class)

    def to_type_parsing_rules_map(self, type_to_property_mapping):
        """Builds a type to parsing rule map."""
        return {
            target_class: self._bind_lookup_account_descriptor_field(property_name, target_class)
            for target_class, property_name in type_to_property_mapping.items()
        }



import yaml

from .AccountDescriptorRepository import AccountDescriptorRepository
from .NodeDescriptorRepository import NodeDescriptorRepository


class BlockchainSettings:
    """Settings describing a blockchain."""

    def __init__(self, settings_dict):
        """Creates blockchain settings from a dictionary."""
        self.blockchain = settings_dict['blockchain']
        self.network = settings_dict['network']
        self.nodes = NodeDescriptorRepository(settings_dict['nodes'])
        self.accounts = AccountDescriptorRepository(settings_dict['accounts'])

    @staticmethod
    def load_from_yaml(yaml_input):
        """Loads settings from YAML."""
        settings_yaml = yaml.load(yaml_input, Loader=yaml.SafeLoader)
        return BlockchainSettings(settings_yaml)



import sha3

from ..CryptoTypes import PrivateKey, PublicKey, Signature
from .external import ed25519


class KeyPair:
    """Represents an ED25519 private and public key."""

    def __init__(self, private_key):
        """Creates a key pair from a private key."""
        self._sk = private_key.bytes[::-1]
        self._pk = ed25519.publickey_hash_unsafe(self._sk, sha3.keccak_512)

    @property
    def public_key(self):
        """Gets the public key."""
        return PublicKey(self._pk)

    @property
    def private_key(self):
        """Gets the private key."""
        return PrivateKey(self._sk[::-1])

    def sign(self, message):
        """Signs a message with the private key."""
        return Signature(ed25519.signature_hash_unsafe(message, self._sk, self._pk, sha3.keccak_512))


class Verifier:
    """Verifies signatures signed by a single key pair."""

    def __init__(self, public_key):
        """Creates a verifier from a public key."""
        self._pk = public_key.bytes

    def verify(self, message, signature):
        """Verifies a message signature."""
        try:
            ed25519.checkvalid_hash(signature.bytes, message, self._pk, sha3.keccak_512)
            return True
        except (ValueError, ed25519.SignatureMismatch):
            return False



import datetime

from ..NetworkTimestamp import NetworkTimestamp as BasicNetworkTimestamp
from ..NetworkTimestamp import NetworkTimestampDatetimeConverter

EPOCH_TIME = datetime.datetime(2021, 3, 16, 0, 6, 25, tzinfo=datetime.timezone.utc)


class NetworkTimestamp(BasicNetworkTimestamp):
    """Represents a symbol network timestamp with millisecond resolution."""

    def add_milliseconds(self, count):
        """Adds a specified number of milliseconds to this timestamp."""
        return NetworkTimestamp(self.timestamp + count)

    def add_seconds(self, count):
        return self.add_milliseconds(1000 * count)

    def to_datetime(self, epoch_time=EPOCH_TIME):
        """Converts this timestamp to a datetime."""
        return NetworkTimestampDatetimeConverter(epoch_time, 'milliseconds').to_datetime(self.timestamp)

    @staticmethod
    def from_datetime(reference_datetime, epoch_time=EPOCH_TIME):
        """Creates a network timestamp from a datetime."""
        return NetworkTimestamp(NetworkTimestampDatetimeConverter(epoch_time, 'milliseconds').to_difference(reference_datetime))



from abc import abstractmethod


class NisTransactionTestDescriptor:
    def __init__(self, transaction_class, transaction_name, transaction_type, network):
        self.transaction_class = transaction_class
        self.transaction_name = transaction_name
        self.transaction_type = transaction_type
        self.network = network


class BasicNisTransactionTest:
    # pylint: disable=no-member

    def test_constants_are_correct(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act + Assert:
        self.assertEqual(test_descriptor.transaction_name, test_descriptor.transaction_class.NAME)
        self.assertEqual(test_descriptor.transaction_type, test_descriptor.transaction_class.TYPE)

    def test_deadline_is_updated_with_timestamp(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()
        transaction = test_descriptor.transaction_class(test_descriptor.network)

        # Act:
        transaction.timestamp = 9876

        # Assert:
        self.assertEqual(9876 + 60 * 60, transaction.deadline)

    @abstractmethod
    def get_test_descriptor(self):
        pass



import os

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519

from .CryptoTypes import PrivateKey


class PrivateKeyStorage:
    """Loads and saves private keys as PEM files with optional encryption in a directory."""

    def __init__(self, directory, password=None):
        """Creates storage for a directory."""
        self.directory = directory
        self.password = None if not password else password.encode('utf8')

    def save(self, name, private_key):
        """Saves a private key with the specified name."""
        wrapped_private_key = ed25519.Ed25519PrivateKey.from_private_bytes(private_key.bytes)
        with open(self._get_file_path(name), 'wb') as outfile:
            encryption_algorithm = serialization.NoEncryption()
            if self.password:
                encryption_algorithm = serialization.BestAvailableEncryption(self.password)

            outfile.write(wrapped_private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=encryption_algorithm))

    def load(self, name):
        """Loads a private key with the specified name."""
        with open(self._get_file_path(name), 'rb') as infile:
            wrapped_private_key = serialization.load_pem_private_key(infile.read(), password=self.password)
            private_key_bytes = wrapped_private_key.private_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PrivateFormat.Raw,
                encryption_algorithm=serialization.NoEncryption())
            return PrivateKey(private_key_bytes)

    def _get_file_path(self, name):
        return os.path.join(self.directory, '{}.pem'.format(name))



from abc import abstractmethod
from binascii import hexlify

from ..BufferWriter import BufferWriter
from ..CryptoTypes import PublicKey
from .NetworkTimestamp import NetworkTimestamp


class Transaction:
    """Represents a transaction."""

    def __init__(self, network, transaction_type):
        """Creates a transaction for the specified network."""
        self.type = transaction_type
        self.version = (network.identifier << 24) + 1
        self.timestamp = 0
        self.signer_public_key = None

    @property
    def deadline(self):
        """Gets the deadline."""
        return NetworkTimestamp(self.timestamp).add_hours(1).timestamp

    @property
    def fee(self):
        """Gets the (minimum) fee."""

    def serialize(self):
        """Serializes this transaction to a buffer."""
        writer = BufferWriter()

        writer.write_int(self.type, 4)
        writer.write_int(self.version, 4)
        writer.write_int(self.timestamp, 4)

        writer.write_int(PublicKey.SIZE, 4)
        writer.write_bytes(self.signer_public_key.bytes)

        writer.write_int(self.fee, 8)
        writer.write_int(self.deadline, 4)

        self.serialize_custom(writer)
        return writer.buffer

    @abstractmethod
    def serialize_custom(self, writer):
        """Serializes custom transaction fields to a writer."""

    @abstractmethod
    def field_names(self):
        """Gets names of custom transaction fields to include in the string representation."""

    class _FieldFormatter:
        def __init__(self):
            self.field_pairs = []

        def add_field(self, key, value):
            self.field_pairs.append((key, self._format_value(value)))

        @staticmethod
        def _format_value(value):
            if isinstance(value, int):
                return '{0} [0x{0:X}]'.format(value)

            if isinstance(value, bytes):
                try:
                    return value.decode('utf8')
                except UnicodeError:
                    return hexlify(value).upper().decode('utf8')

            return value

        def __str__(self):
            max_name_length = max(len(pair[0]) for pair in self.field_pairs)
            formatted_fields = ['{} = {}'.format(pair[0].rjust(max_name_length, ' '), pair[1]) for pair in self.field_pairs]
            return '\n'.join(formatted_fields)

    def __str__(self):
        formatter = Transaction._FieldFormatter()
        for name in ['type', 'version', 'timestamp', 'signer_public_key', 'fee', 'deadline'] + self.field_names():
            formatter.add_field(name, getattr(self, name))

        return str(formatter)



from abc import abstractmethod


class NetworkTestDescriptor:
    def __init__(
            self,
            network_class,
            deterministic_public_key,
            expected_mainnet_address_network_pair,
            expected_testnet_address_network_pair):
        # pylint: disable=too-many-function-args

        self.network_class = network_class
        self.deterministic_public_key = deterministic_public_key
        (self.expected_mainnet_address, self.mainnet_network) = expected_mainnet_address_network_pair
        (self.expected_testnet_address, self.testnet_network) = expected_testnet_address_network_pair


class BasicNetworkTest:
    # pylint: disable=no-member

    def test_can_convert_mainnet_public_key_to_address(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act:
        address = test_descriptor.mainnet_network.public_key_to_address(test_descriptor.deterministic_public_key)

        # Assert:
        self.assertEqual(test_descriptor.expected_mainnet_address, address)

    def test_can_convert_testnet_public_key_to_address(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act:
        address = test_descriptor.testnet_network.public_key_to_address(test_descriptor.deterministic_public_key)

        # Assert:
        self.assertEqual(test_descriptor.expected_testnet_address, address)

    def _assert_network(self, network, expected_name, expected_identifier):
        self.assertEqual(expected_name, network.name)
        self.assertEqual(expected_identifier, network.identifier)

    @abstractmethod
    def get_test_descriptor(self):
        pass



import unittest

from symbolchain.core.BufferWriter import BufferWriter


class BufferWriterTest(unittest.TestCase):
    def test_output_buffer_is_initially_empty(self):
        # Act:
        writer = BufferWriter()

        # Assert:
        self.assertEqual(bytes(), writer.buffer)

    def _assert_can_write_int_endianness_little(self, writer):
        # Act:
        writer.write_int(0x05010203, 4)
        writer.write_int(0x00FF, 2)
        writer.write_int(0xEE00, 2)

        # Assert:
        self.assertEqual(bytes([0x03, 0x02, 0x01, 0x05, 0xFF, 0x00, 0x00, 0xEE]), writer.buffer)

    def test_can_write_int_endianness_default(self):
        self._assert_can_write_int_endianness_little(BufferWriter())

    def test_can_write_int_endianness_little(self):
        self._assert_can_write_int_endianness_little(BufferWriter('little'))

    def test_can_write_int_endianness_big(self):
        # Arrange:
        writer = BufferWriter('big')

        # Act:
        writer.write_int(0x05010203, 4)
        writer.write_int(0x00FF, 2)
        writer.write_int(0xEE00, 2)

        # Assert:
        self.assertEqual(bytes([0x05, 0x01, 0x02, 0x03, 0x00, 0xFF, 0xEE, 0x00]), writer.buffer)

    def test_can_write_string(self):
        # Arrange:
        writer = BufferWriter()

        # Act:
        writer.write_string('bar')

        # Assert:
        self.assertEqual(bytes([0x62, 0x61, 0x72]), writer.buffer)

    def test_can_write_hex_string(self):
        # Arrange:
        writer = BufferWriter()

        # Act:
        writer.write_hex_string('C0AA0008')

        # Assert:
        self.assertEqual(bytes([0xC0, 0xAA, 0x00, 0x08]), writer.buffer)

    def test_can_write_bytes(self):
        # Arrange:
        writer = BufferWriter()

        # Act:
        writer.write_bytes(bytes([0x99, 0x00, 0x40, 0x30, 0x78]))

        # Assert:
        self.assertEqual(bytes([0x99, 0x00, 0x40, 0x30, 0x78]), writer.buffer)

    def _assert_can_write_all_endianness_little(self, writer):
        # Act:
        writer.write_string('bar')
        writer.write_int(0x05010203, 4)
        writer.write_bytes(bytes([0x99, 0x00, 0x40, 0x30, 0x78]))
        writer.write_hex_string('C0AA0008')

        # Assert:
        expected_bytes = bytes([
            0x62, 0x61, 0x72,
            0x03, 0x02, 0x01, 0x05,
            0x99, 0x00, 0x40, 0x30, 0x78,
            0xC0, 0xAA, 0x00, 0x08
        ])
        self.assertEqual(expected_bytes, writer.buffer)

    def test_can_write_all_endianness_default(self):
        self._assert_can_write_all_endianness_little(BufferWriter())

    def test_can_write_all_endianness_little(self):
        self._assert_can_write_all_endianness_little(BufferWriter('little'))

    def test_can_write_all_endianness_big(self):
        # Arrange:
        writer = BufferWriter('big')

        # Act:
        writer.write_string('bar')
        writer.write_int(0x05010203, 4)
        writer.write_bytes(bytes([0x99, 0x00, 0x40, 0x30, 0x78]))
        writer.write_hex_string('C0AA0008')

        # Assert:
        expected_bytes = bytes([
            0x62, 0x61, 0x72,
            0x05, 0x01, 0x02, 0x03,
            0x99, 0x00, 0x40, 0x30, 0x78,
            0xC0, 0xAA, 0x00, 0x08
        ])
        self.assertEqual(expected_bytes, writer.buffer)



import os
import unittest
from binascii import hexlify, unhexlify
from collections import namedtuple

import yaml

from symbolchain.core.BufferReader import BufferReader
from symbolchain.core.CryptoTypes import PrivateKey, PublicKey, Signature
from symbolchain.core.sym.KeyPair import KeyPair, Verifier
from symbolchain.core.sym.VotingKeysGenerator import VotingKeysGenerator


class VotingKeysGeneratorTest(unittest.TestCase):
    # region private key generators

    class SeededPrivateKeyGenerator:
        def __init__(self, values):
            self.values = values
            self.next_index = 0

        def generate(self):
            self.next_index += 1
            return self.values[self.next_index - 1]

    class FibPrivateKeyGenerator:
        def __init__(self, fill_private_key=False):
            self.fill_private_key = fill_private_key
            self.value1 = 1
            self.value2 = 2

        def generate(self):
            next_value = self.value1 + self.value2
            self.value1 = self.value2
            self.value2 = next_value

            seed_value = next_value % 256

            if not self.fill_private_key:
                return PrivateKey(seed_value.to_bytes(PrivateKey.SIZE, 'big'))

            return PrivateKey(bytes([(seed_value + i) % 256 for i in range(0, PrivateKey.SIZE)]))

    # endregion

    # region basic test

    def test_can_generate_header(self):
        # Arrange:
        root_key_pair = KeyPair(PrivateKey.random())
        voting_keys_generator = VotingKeysGenerator(root_key_pair)

        # Act:
        voting_keys_buffer = voting_keys_generator.generate(7, 11)

        # Assert:
        self.assertEqual(32 + PublicKey.SIZE + 16 + 5 * (PrivateKey.SIZE + Signature.SIZE), len(voting_keys_buffer))

        reader = BufferReader(voting_keys_buffer)
        self.assertEqual(7, reader.read_int(8))
        self.assertEqual(11, reader.read_int(8))
        self.assertEqual(0xFFFFFFFFFFFFFFFF, reader.read_int(8))
        self.assertEqual(0xFFFFFFFFFFFFFFFF, reader.read_int(8))

        self.assertEqual(root_key_pair.public_key, PublicKey(reader.read_bytes(PublicKey.SIZE)))
        self.assertEqual(7, reader.read_int(8))
        self.assertEqual(11, reader.read_int(8))

    def test_can_generate_random_child_keys(self):
        # Arrange:
        root_key_pair = KeyPair(PrivateKey.random())
        voting_keys_generator = VotingKeysGenerator(root_key_pair)

        # Act:
        voting_keys_buffer = voting_keys_generator.generate(7, 11)

        # Assert:
        self.assertEqual(32 + PublicKey.SIZE + 16 + 5 * (PrivateKey.SIZE + Signature.SIZE), len(voting_keys_buffer))

        reader = BufferReader(voting_keys_buffer)
        reader.read_bytes(32 + PublicKey.SIZE + 16)  # skip header

        verifier = Verifier(root_key_pair.public_key)
        for i in reversed(range(0, 5)):
            child_private_key = PrivateKey(reader.read_bytes(PrivateKey.SIZE))
            signature = Signature(reader.read_bytes(Signature.SIZE))

            child_key_pair = KeyPair(child_private_key)
            signed_payload = child_key_pair.public_key.bytes + (7 + i).to_bytes(8, 'little')

            self.assertTrue(verifier.verify(signed_payload, signature), 'child at {}'.format(i))

    # endregion

    # region test vectors

    @staticmethod
    def _print_side_by_side_buffers_diff(expected_buffer, actual_buffer, batch_size):
        for i in range(0, len(expected_buffer) // batch_size):
            part_start_offset = i * batch_size
            part_end_offset = (i + 1) * batch_size
            expected_part = hexlify(expected_buffer[part_start_offset:part_end_offset]).decode('utf8').upper()
            actual_part = hexlify(actual_buffer[part_start_offset:part_end_offset]).decode('utf8').upper()
            print('E {} {} A {}'.format(expected_part, '==' if expected_part == actual_part else '!=', actual_part))

    def _run_test_vector(self, name, private_key_generator):
        # Arrange:
        test_vectors_filename = os.path.join(os.path.dirname(__file__), 'resources', 'voting_keys_generator_test_vectors.yaml')
        with open(test_vectors_filename, 'rt') as infile:
            test_vectors_yaml = yaml.load(infile.read(), Loader=yaml.SafeLoader)
            test_vector_yaml = next(test_vector_yaml for test_vector_yaml in test_vectors_yaml if name == test_vector_yaml['name'])
            test_vector = namedtuple('TestVector', test_vector_yaml.keys())(*test_vector_yaml.values())

            root_private_key = PrivateKey(test_vector.root_private_key)
            voting_keys_generator = VotingKeysGenerator(KeyPair(root_private_key), private_key_generator)

            # Act:
            voting_keys_buffer = voting_keys_generator.generate(test_vector.start_epoch, test_vector.end_epoch)

            # Assert:
            expected_voting_keys_buffer = unhexlify(test_vector.expected_file_hex)

            if expected_voting_keys_buffer != voting_keys_buffer:
                self._print_side_by_side_buffers_diff(expected_voting_keys_buffer, voting_keys_buffer, 16)

            self.assertEqual(expected_voting_keys_buffer, voting_keys_buffer)

    def test_can_generate_test_vector_1(self):
        self._run_test_vector('test_vector_1', self.FibPrivateKeyGenerator().generate)

    def test_can_generate_test_vector_2(self):
        self._run_test_vector('test_vector_2', self.FibPrivateKeyGenerator(True).generate)

    def test_can_generate_test_vector_3(self):
        # Arrange:
        private_key_generator = self.SeededPrivateKeyGenerator([
            PrivateKey('12F98B7CB64A6D840931A2B624FB1EACAFA2C25C3EF0018CD67E8D470A248B2F'),
            PrivateKey('B5593870940F28DAEE262B26367B69143AD85E43048D23E624F4ED8008C0427F'),
            PrivateKey('6CFC879ABCCA78F5A4C9739852C7C643AEC3990E93BF4C6F685EB58224B16A59')
        ])

        # Act + Assert:
        self._run_test_vector('test_vector_3', private_key_generator.generate)

    # endregion



import datetime
from abc import abstractmethod


class NetworkTimestamp:
    """Represents a network timestamp."""

    def __init__(self, timestamp):
        """Creates a timestamp."""
        self.timestamp = timestamp

    @property
    def epochal(self):
        """Determines if this is the epochal timestamp."""
        return 0 == self.timestamp

    @abstractmethod
    def add_seconds(self, count):
        """Adds a specified number of seconds to this timestamp."""

    def add_minutes(self, count):
        """Adds a specified number of minutes to this timestamp."""
        return self.add_seconds(60 * count)

    def add_hours(self, count):
        """Adds a specified number of hours to this timestamp."""
        return self.add_minutes(60 * count)

    def __eq__(self, other):
        return isinstance(other, NetworkTimestamp) and self.timestamp == other.timestamp

    def __str__(self):
        return str(self.timestamp)


class NetworkTimestampDatetimeConverter:
    """Provides utilities for converting between network timestamps and datetimes."""

    def __init__(self, epoch, time_units):
        """Creates a converter given an epoch and base time units."""
        self.epoch = epoch
        self.time_units = time_units

    def to_datetime(self, raw_timestamp):
        """Converts a network timestamp to a datetime."""
        return self.epoch + datetime.timedelta(**{self.time_units: raw_timestamp})

    def to_difference(self, reference_datetime):
        """Creates a network timestamp from a datetime."""
        if reference_datetime < self.epoch:
            raise ValueError('timestamp cannot be before epoch')

        return (reference_datetime - self.epoch) / datetime.timedelta(**{self.time_units: 1})



class TransactionDescriptorProcessor:
    """Processes and looks up transaction descriptor properties."""

    def __init__(self, transaction_descriptor, type_parsing_rules):
        """Creates a transaction descriptor processor."""
        self.transaction_descriptor = transaction_descriptor
        self.type_parsing_rules = type_parsing_rules or {}
        self.type_hints = {}

    def lookup_value(self, key):
        """Looks up the value for key."""
        if key not in self.transaction_descriptor:
            raise ValueError('transaction descriptor does not have attribute {}'.format(key))

        value = self.transaction_descriptor[key]

        type_hint = self.type_hints.get(key)
        if type_hint in self.type_parsing_rules:
            value = self.type_parsing_rules[type_hint](value)

        return value

    def copy_to(self, transaction, ignore_keys=None):
        """Copies all descriptor information to a transaction."""
        for key in self.transaction_descriptor.keys():
            if ignore_keys and key in ignore_keys:
                continue

            if not hasattr(transaction, key):
                raise ValueError('transaction does not have attribute {}'.format(key))

            value = self.lookup_value(key)
            if not any(isinstance(value, type_class) for type_class in [str, bytes, tuple]) and hasattr(value, '__iter__'):
                for item in value:
                    getattr(transaction, key).append(item)
            else:
                setattr(transaction, key, value)

    def set_type_hints(self, type_hints):
        """Sets type hints."""
        self.type_hints = type_hints or {}
